"""Training and validation Lightning data modules."""

from pathlib import Path
from typing import Any, Callable, Literal, Optional, Union
from warnings import warn

import numpy as np
import pytorch_lightning as L
from numpy.typing import NDArray
from torch.utils.data import DataLoader, IterableDataset

from careamics.config import DataConfig
from careamics.config.support import SupportedData
from careamics.config.transformations import TransformModel
from careamics.dataset.dataset_utils import (
    get_files_size,
    list_files,
    validate_source_target_files,
)
from careamics.dataset.in_memory_dataset import (
    InMemoryDataset,
)
from careamics.dataset.iterable_dataset import (
    PathIterableDataset,
)
from careamics.file_io.read import get_read_func
from careamics.utils import get_logger, get_ram_size

DatasetType = Union[InMemoryDataset, PathIterableDataset]

logger = get_logger(__name__)


class TrainDataModule(L.LightningDataModule):
    """
    CAREamics Ligthning training and validation data module.

    The data module can be used with Path, str or numpy arrays. In the case of
    numpy arrays, it loads and computes all the patches in memory. For Path and str
    inputs, it calculates the total file size and estimate whether it can fit in
    memory. If it does not, it iterates through the files. This behaviour can be
    deactivated by setting `use_in_memory` to False, in which case it will
    always use the iterating dataset to train on a Path or str.

    The data can be either a folder containing images or a single file.

    Validation can be omitted, in which case the validation data is extracted from
    the training data. The percentage of the training data to use for validation,
    as well as the minimum number of patches or files to split from the training
    data can be set using `val_percentage` and `val_minimum_split`, respectively.

    To read custom data types, you can set `data_type` to `custom` in `data_config`
    and provide a function that returns a numpy array from a path as
    `read_source_func` parameter. The function will receive a Path object and
    an axies string as arguments, the axes being derived from the `data_config`.

    You can also provide a `fnmatch` and `Path.rglob` compatible expression (e.g.
    "*.czi") to filter the files extension using `extension_filter`.

    Parameters
    ----------
    data_config : DataModel
        Pydantic model for CAREamics data configuration.
    train_data : pathlib.Path or str or numpy.ndarray
        Training data, can be a path to a folder, a file or a numpy array.
    val_data : pathlib.Path or str or numpy.ndarray, optional
        Validation data, can be a path to a folder, a file or a numpy array, by
        default None.
    train_data_target : pathlib.Path or str or numpy.ndarray, optional
        Training target data, can be a path to a folder, a file or a numpy array, by
        default None.
    val_data_target : pathlib.Path or str or numpy.ndarray, optional
        Validation target data, can be a path to a folder, a file or a numpy array,
        by default None.
    read_source_func : Callable, optional
        Function to read the source data, by default None. Only used for `custom`
        data type (see DataModel).
    extension_filter : str, optional
        Filter for file extensions, by default "". Only used for `custom` data types
        (see DataModel).
    val_percentage : float, optional
        Percentage of the training data to use for validation, by default 0.1. Only
        used if `val_data` is None.
    val_minimum_split : int, optional
        Minimum number of patches or files to split from the training data for
        validation, by default 5. Only used if `val_data` is None.
    use_in_memory : bool, optional
        Use in memory dataset if possible, by default True.

    Attributes
    ----------
    data_config : DataModel
        CAREamics data configuration.
    data_type : SupportedData
        Expected data type, one of "tiff", "array" or "custom".
    batch_size : int
        Batch size.
    use_in_memory : bool
        Whether to use in memory dataset if possible.
    train_data : pathlib.Path or numpy.ndarray
        Training data.
    val_data : pathlib.Path or numpy.ndarray
        Validation data.
    train_data_target : pathlib.Path or numpy.ndarray
        Training target data.
    val_data_target : pathlib.Path or numpy.ndarray
        Validation target data.
    val_percentage : float
        Percentage of the training data to use for validation, if no validation data is
        provided.
    val_minimum_split : int
        Minimum number of patches or files to split from the training data for
        validation, if no validation data is provided.
    read_source_func : Optional[Callable]
        Function to read the source data, used if `data_type` is `custom`.
    extension_filter : str
        Filter for file extensions, used if `data_type` is `custom`.
    """

    def __init__(
        self,
        data_config: DataConfig,
        train_data: Union[Path, str, NDArray],
        val_data: Optional[Union[Path, str, NDArray]] = None,
        train_data_target: Optional[Union[Path, str, NDArray]] = None,
        val_data_target: Optional[Union[Path, str, NDArray]] = None,
        read_source_func: Optional[Callable] = None,
        extension_filter: str = "",
        val_percentage: float = 0.1,
        val_minimum_split: int = 5,
        use_in_memory: bool = True,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        data_config : DataModel
            Pydantic model for CAREamics data configuration.
        train_data : pathlib.Path or str or numpy.ndarray
            Training data, can be a path to a folder, a file or a numpy array.
        val_data : pathlib.Path or str or numpy.ndarray, optional
            Validation data, can be a path to a folder, a file or a numpy array, by
            default None.
        train_data_target : pathlib.Path or str or numpy.ndarray, optional
            Training target data, can be a path to a folder, a file or a numpy array, by
            default None.
        val_data_target : pathlib.Path or str or numpy.ndarray, optional
            Validation target data, can be a path to a folder, a file or a numpy array,
            by default None.
        read_source_func : Callable, optional
            Function to read the source data, by default None. Only used for `custom`
            data type (see DataModel).
        extension_filter : str, optional
            Filter for file extensions, by default "". Only used for `custom` data types
            (see DataModel).
        val_percentage : float, optional
            Percentage of the training data to use for validation, by default 0.1. Only
            used if `val_data` is None.
        val_minimum_split : int, optional
            Minimum number of patches or files to split from the training data for
            validation, by default 5. Only used if `val_data` is None.
        use_in_memory : bool, optional
            Use in memory dataset if possible, by default True.

        Raises
        ------
        NotImplementedError
            Raised if target data is provided.
        ValueError
            If the input types are mixed (e.g. Path and numpy.ndarray).
        ValueError
            If the data type is `custom` and no `read_source_func` is provided.
        ValueError
            If the data type is `array` and the input is not a numpy array.
        ValueError
            If the data type is `tiff` and the input is neither a Path nor a str.
        """
        super().__init__()

        # check input types coherence (no mixed types)
        inputs = [train_data, val_data, train_data_target, val_data_target]
        types_set = {type(i) for i in inputs}
        if len(types_set) > 2:  # None + expected type
            raise ValueError(
                f"Inputs for `train_data`, `val_data`, `train_data_target` and "
                f"`val_data_target` must be of the same type or None. Got "
                f"{types_set}."
            )

        # check that a read source function is provided for custom types
        if data_config.data_type == SupportedData.CUSTOM and read_source_func is None:
            raise ValueError(
                f"Data type {SupportedData.CUSTOM} is not allowed without "
                f"specifying a `read_source_func` and an `extension_filer`."
            )

        # check correct input type
        if (
            isinstance(train_data, np.ndarray)
            and data_config.data_type != SupportedData.ARRAY
        ):
            raise ValueError(
                f"Received a numpy array as input, but the data type was set to "
                f"{data_config.data_type}. Set the data type in the configuration "
                f"to {SupportedData.ARRAY} to train on numpy arrays."
            )

        # and that Path or str are passed, if tiff file type specified
        elif (isinstance(train_data, Path) or isinstance(train_data, str)) and (
            data_config.data_type != SupportedData.TIFF
            and data_config.data_type != SupportedData.CUSTOM
        ):
            raise ValueError(
                f"Received a path as input, but the data type was neither set to "
                f"{SupportedData.TIFF} nor {SupportedData.CUSTOM}. Set the data type "
                f"in the configuration to {SupportedData.TIFF} or "
                f"{SupportedData.CUSTOM} to train on files."
            )

        # configuration
        self.data_config: DataConfig = data_config
        self.data_type: str = data_config.data_type
        self.batch_size: int = data_config.batch_size
        self.use_in_memory: bool = use_in_memory

        # data: make data Path or np.ndarray, use type annotations for mypy
        self.train_data: Union[Path, NDArray] = (
            Path(train_data) if isinstance(train_data, str) else train_data
        )

        self.val_data: Union[Path, NDArray] = (
            Path(val_data) if isinstance(val_data, str) else val_data
        )

        self.train_data_target: Union[Path, NDArray] = (
            Path(train_data_target)
            if isinstance(train_data_target, str)
            else train_data_target
        )

        self.val_data_target: Union[Path, NDArray] = (
            Path(val_data_target)
            if isinstance(val_data_target, str)
            else val_data_target
        )

        # validation split
        self.val_percentage = val_percentage
        self.val_minimum_split = val_minimum_split

        # read source function corresponding to the requested type
        if data_config.data_type == SupportedData.CUSTOM.value:
            # mypy check
            assert read_source_func is not None

            self.read_source_func: Callable = read_source_func

        elif data_config.data_type != SupportedData.ARRAY:
            self.read_source_func = get_read_func(data_config.data_type)

        self.extension_filter: str = extension_filter

        # Pytorch dataloader parameters
        self.dataloader_params: dict[str, Any] = (
            data_config.dataloader_params if data_config.dataloader_params else {}
        )

    def prepare_data(self) -> None:
        """
        Hook used to prepare the data before calling `setup`.

        Here, we only need to examine the data if it was provided as a str or a Path.

        TODO: from lightning doc:
        prepare_data is called from the main process. It is not recommended to assign
        state here (e.g. self.x = y) since it is called on a single process and if you
        assign states here then they won't be available for other processes.

        https://lightning.ai/docs/pytorch/stable/data/datamodule.html
        """
        # if the data is a Path or a str
        if (
            not isinstance(self.train_data, np.ndarray)
            and not isinstance(self.val_data, np.ndarray)
            and not isinstance(self.train_data_target, np.ndarray)
            and not isinstance(self.val_data_target, np.ndarray)
        ):
            # list training files
            self.train_files = list_files(
                self.train_data, self.data_type, self.extension_filter
            )
            self.train_files_size = get_files_size(self.train_files)

            # list validation files
            if self.val_data is not None:
                self.val_files = list_files(
                    self.val_data, self.data_type, self.extension_filter
                )

            # same for target data
            if self.train_data_target is not None:
                self.train_target_files: list[Path] = list_files(
                    self.train_data_target, self.data_type, self.extension_filter
                )

                # verify that they match the training data
                validate_source_target_files(self.train_files, self.train_target_files)

            if self.val_data_target is not None:
                self.val_target_files = list_files(
                    self.val_data_target, self.data_type, self.extension_filter
                )

                # verify that they match the validation data
                validate_source_target_files(self.val_files, self.val_target_files)

    def setup(self, *args: Any, **kwargs: Any) -> None:
        """Hook called at the beginning of fit, validate, or predict.

        Parameters
        ----------
        *args : Any
            Unused.
        **kwargs : Any
            Unused.
        """
        # if numpy array
        if self.data_type == SupportedData.ARRAY:
            # mypy checks
            assert isinstance(self.train_data, np.ndarray)
            if self.train_data_target is not None:
                assert isinstance(self.train_data_target, np.ndarray)

            # train dataset
            self.train_dataset: DatasetType = InMemoryDataset(
                data_config=self.data_config,
                inputs=self.train_data,
                input_target=self.train_data_target,
            )

            # validation dataset
            if self.val_data is not None:
                # mypy checks
                assert isinstance(self.val_data, np.ndarray)
                if self.val_data_target is not None:
                    assert isinstance(self.val_data_target, np.ndarray)

                # create its own dataset
                self.val_dataset: DatasetType = InMemoryDataset(
                    data_config=self.data_config,
                    inputs=self.val_data,
                    input_target=self.val_data_target,
                )
            else:
                # extract validation from the training patches
                self.val_dataset = self.train_dataset.split_dataset(
                    percentage=self.val_percentage,
                    minimum_patches=self.val_minimum_split,
                )

        # else we read files
        else:
            # Heuristics, if the file size is smaller than 80% of the RAM,
            # we run the training in memory, otherwise we switch to iterable dataset
            # The switch is deactivated if use_in_memory is False
            if self.use_in_memory and self.train_files_size < get_ram_size() * 0.8:
                # train dataset
                self.train_dataset = InMemoryDataset(
                    data_config=self.data_config,
                    inputs=self.train_files,
                    input_target=(
                        self.train_target_files if self.train_data_target else None
                    ),
                    read_source_func=self.read_source_func,
                )

                # validation dataset
                if self.val_data is not None:
                    self.val_dataset = InMemoryDataset(
                        data_config=self.data_config,
                        inputs=self.val_files,
                        input_target=(
                            self.val_target_files if self.val_data_target else None
                        ),
                        read_source_func=self.read_source_func,
                    )
                else:
                    # split dataset
                    self.val_dataset = self.train_dataset.split_dataset(
                        percentage=self.val_percentage,
                        minimum_patches=self.val_minimum_split,
                    )

            # else if the data is too large, load file by file during training
            else:
                # create training dataset
                self.train_dataset = PathIterableDataset(
                    data_config=self.data_config,
                    src_files=self.train_files,
                    target_files=(
                        self.train_target_files if self.train_data_target else None
                    ),
                    read_source_func=self.read_source_func,
                )

                # create validation dataset
                if self.val_data is not None:
                    # create its own dataset
                    self.val_dataset = PathIterableDataset(
                        data_config=self.data_config,
                        src_files=self.val_files,
                        target_files=(
                            self.val_target_files if self.val_data_target else None
                        ),
                        read_source_func=self.read_source_func,
                    )
                elif len(self.train_files) <= self.val_minimum_split:
                    raise ValueError(
                        f"Not enough files to split a minimum of "
                        f"{self.val_minimum_split} files, got {len(self.train_files)} "
                        f"files."
                    )
                else:
                    # extract validation from the training patches
                    self.val_dataset = self.train_dataset.split_dataset(
                        percentage=self.val_percentage,
                        minimum_number=self.val_minimum_split,
                    )

    def get_data_statistics(self) -> tuple[list[float], list[float]]:
        """Return training data statistics.

        Returns
        -------
        tuple of list
            Means and standard deviations across channels of the training data.
        """
        return self.train_dataset.get_data_statistics()

    def train_dataloader(self) -> Any:
        """
        Create a dataloader for training.

        Returns
        -------
        Any
            Training dataloader.
        """
        # check because iterable dataset cannot be shuffled
        if not isinstance(self.train_dataset, IterableDataset):
            if ("shuffle" in self.dataloader_params) and (
                not self.dataloader_params["shuffle"]
            ):
                warn(
                    "Dataloader parameters include `shuffle=False`, this will be "
                    "passed to the training dataloader and may result in bad results.",
                    stacklevel=1,
                )
            else:
                self.dataloader_params["shuffle"] = True

        return DataLoader(
            self.train_dataset, batch_size=self.batch_size, **self.dataloader_params
        )

    def val_dataloader(self) -> Any:
        """
        Create a dataloader for validation.

        Returns
        -------
        Any
            Validation dataloader.
        """
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
        )


def create_train_datamodule(
    train_data: Union[str, Path, NDArray],
    data_type: Union[Literal["array", "tiff", "custom"], SupportedData],
    patch_size: list[int],
    axes: str,
    batch_size: int,
    val_data: Optional[Union[str, Path, NDArray]] = None,
    transforms: Optional[list[TransformModel]] = None,
    train_target_data: Optional[Union[str, Path, NDArray]] = None,
    val_target_data: Optional[Union[str, Path, NDArray]] = None,
    read_source_func: Optional[Callable] = None,
    extension_filter: str = "",
    val_percentage: float = 0.1,
    val_minimum_patches: int = 5,
    dataloader_params: Optional[dict] = None,
    use_in_memory: bool = True,
    use_n2v2: bool = False,
    struct_n2v_axis: Literal["horizontal", "vertical", "none"] = "none",
    struct_n2v_span: int = 5,
) -> TrainDataModule:
    """Create a TrainDataModule.

    This function is used to explicitly pass the parameters usually contained in a
    `data_model` configuration to a TrainDataModule.

    Since the lightning datamodule has no access to the model, make sure that the
    parameters passed to the datamodule are consistent with the model's requirements and
    are coherent.

    The data module can be used with Path, str or numpy arrays. In the case of
    numpy arrays, it loads and computes all the patches in memory. For Path and str
    inputs, it calculates the total file size and estimate whether it can fit in
    memory. If it does not, it iterates through the files. This behaviour can be
    deactivated by setting `use_in_memory` to False, in which case it will
    always use the iterating dataset to train on a Path or str.

    To use array data, set `data_type` to `array` and pass a numpy array to
    `train_data`.

    In particular, N2V requires a specific transformation (N2V manipulates), which is
    not compatible with supervised training. The default transformations applied to the
    training patches are defined in `careamics.config.data_model`. To use different
    transformations, pass a list of transforms. See examples for more details.

    By default, CAREamics only supports types defined in
    `careamics.config.support.SupportedData`. To read custom data types, you can set
    `data_type` to `custom` and provide a function that returns a numpy array from a
    path. Additionally, pass a `fnmatch` and `Path.rglob` compatible expression (e.g.
    "*.jpeg") to filter the files extension using `extension_filter`.

    In the absence of validation data, the validation data is extracted from the
    training data. The percentage of the training data to use for validation, as well as
    the minimum number of patches to split from the training data for validation can be
    set using `val_percentage` and `val_minimum_patches`, respectively.

    In `dataloader_params`, you can pass any parameter accepted by PyTorch dataloaders,
    except for `batch_size`, which is set by the `batch_size` parameter.

    Finally, if you intend to use N2V family of algorithms, you can set `use_n2v2` to
    use N2V2, and set the `struct_n2v_axis` and `struct_n2v_span` parameters to define
    the axis and span of the structN2V mask. These parameters are without effect if
    a `train_target_data` or if `transforms` are provided.

    Parameters
    ----------
    train_data : pathlib.Path or str or numpy.ndarray
        Training data.
    data_type : {"array", "tiff", "custom"}
        Data type, see `SupportedData` for available options.
    patch_size : list of int
        Patch size, 2D or 3D patch size.
    axes : str
        Axes of the data, chosen amongst SCZYX.
    batch_size : int
        Batch size.
    val_data : pathlib.Path or str or numpy.ndarray, optional
        Validation data, by default None.
    transforms : list of Transforms, optional
        List of transforms to apply to training patches. If None, default transforms
        are applied.
    train_target_data : pathlib.Path or str or numpy.ndarray, optional
        Training target data, by default None.
    val_target_data : pathlib.Path or str or numpy.ndarray, optional
        Validation target data, by default None.
    read_source_func : Callable, optional
        Function to read the source data, used if `data_type` is `custom`, by
        default None.
    extension_filter : str, optional
        Filter for file extensions, used if `data_type` is `custom`, by default "".
    val_percentage : float, optional
        Percentage of the training data to use for validation if no validation data
        is given, by default 0.1.
    val_minimum_patches : int, optional
        Minimum number of patches to split from the training data for validation if
        no validation data is given, by default 5.
    dataloader_params : dict, optional
        Pytorch dataloader parameters, by default {}.
    use_in_memory : bool, optional
        Use in memory dataset if possible, by default True.
    use_n2v2 : bool, optional
        Use N2V2 transformation during training, by default False.
    struct_n2v_axis : {"horizontal", "vertical", "none"}, optional
        Axis for the structN2V mask, only applied if `struct_n2v_axis` is `none`, by
        default "none".
    struct_n2v_span : int, optional
        Span for the structN2V mask, by default 5.

    Returns
    -------
    TrainDataModule
        CAREamics training Lightning data module.

    Examples
    --------
    Create a TrainingDataModule with default transforms with a numpy array:
    >>> import numpy as np
    >>> from careamics.lightning import create_train_datamodule
    >>> my_array = np.arange(256).reshape(16, 16)
    >>> data_module = create_train_datamodule(
    ...     train_data=my_array,
    ...     data_type="array",
    ...     patch_size=(8, 8),
    ...     axes='YX',
    ...     batch_size=2,
    ... )

    For custom data types (those not supported by CAREamics), then one can pass a read
    function and a filter for the files extension:
    >>> import numpy as np
    >>> from careamics.lightning import create_train_datamodule
    >>>
    >>> def read_npy(path):
    ...     return np.load(path)
    >>>
    >>> data_module = create_train_datamodule(
    ...     train_data="path/to/data",
    ...     data_type="custom",
    ...     patch_size=(8, 8),
    ...     axes='YX',
    ...     batch_size=2,
    ...     read_source_func=read_npy,
    ...     extension_filter="*.npy",
    ... )

    If you want to use a different set of transformations, you can pass a list of
    transforms:
    >>> import numpy as np
    >>> from careamics.lightning import create_train_datamodule
    >>> from careamics.config.support import SupportedTransform
    >>> my_array = np.arange(256).reshape(16, 16)
    >>> my_transforms = [
    ...     {
    ...         "name": SupportedTransform.XY_FLIP.value,
    ...     }
    ... ]
    >>> data_module = create_train_datamodule(
    ...     train_data=my_array,
    ...     data_type="array",
    ...     patch_size=(8, 8),
    ...     axes='YX',
    ...     batch_size=2,
    ...     transforms=my_transforms,
    ... )
    """
    if dataloader_params is None:
        dataloader_params = {}

    data_dict: dict[str, Any] = {
        "mode": "train",
        "data_type": data_type,
        "patch_size": patch_size,
        "axes": axes,
        "batch_size": batch_size,
        "dataloader_params": dataloader_params,
    }

    # if transforms are passed (otherwise it will use the default ones)
    if transforms is not None:
        data_dict["transforms"] = transforms

    # validate configuration
    data_config = DataConfig(**data_dict)

    # N2V specific checks, N2V, structN2V, and transforms
    if data_config.has_n2v_manipulate():
        # there is not target, n2v2 and structN2V can be changed
        if train_target_data is None:
            data_config.set_N2V2(use_n2v2)
            data_config.set_structN2V_mask(struct_n2v_axis, struct_n2v_span)
        else:
            raise ValueError(
                "Cannot have both supervised training (target data) and "
                "N2V manipulation in the transforms. Pass a list of transforms "
                "that is compatible with your supervised training."
            )

    # sanity check on the dataloader parameters
    if "batch_size" in dataloader_params:
        # remove it
        del dataloader_params["batch_size"]

    return TrainDataModule(
        data_config=data_config,
        train_data=train_data,
        val_data=val_data,
        train_data_target=train_target_data,
        val_data_target=val_target_data,
        read_source_func=read_source_func,
        extension_filter=extension_filter,
        val_percentage=val_percentage,
        val_minimum_split=val_minimum_patches,
        use_in_memory=use_in_memory,
    )



"""
Utility functions needed by dataloader & co.
"""

from typing import List

import numpy as np
from skimage.io import imread, imsave


def load_tiff(path):
    """
    Returns a 4d numpy array: num_imgs*h*w*num_channels
    """
    data = imread(path, plugin="tifffile")
    return data


def save_tiff(path, data):
    imsave(path, data, plugin="tifffile")


def load_tiffs(paths):
    data = [load_tiff(path) for path in paths]
    return np.concatenate(data, axis=0)


def split_in_half(s, e):
    n = e - s
    s1 = list(np.arange(n // 2))
    s2 = list(np.arange(n // 2, n))
    return [x + s for x in s1], [x + s for x in s2]


def adjust_for_imbalance_in_fraction_value(
    val: List[int],
    test: List[int],
    val_fraction: float,
    test_fraction: float,
    total_size: int,
):
    """
    here, val and test are divided almost equally. Here, we need to take into account their respective fractions
    and pick elements rendomly from one array and put in the other array.
    """
    if val_fraction == 0:
        test += val
        val = []
    elif test_fraction == 0:
        val += test
        test = []
    else:
        diff_fraction = test_fraction - val_fraction
        if diff_fraction > 0:
            imb_count = int(diff_fraction * total_size / 2)
            val = list(np.random.RandomState(seed=955).permutation(val))
            test += val[:imb_count]
            val = val[imb_count:]
        elif diff_fraction < 0:
            imb_count = int(-1 * diff_fraction * total_size / 2)
            test = list(np.random.RandomState(seed=955).permutation(test))
            val += test[:imb_count]
            test = test[imb_count:]
    return val, test


def get_datasplit_tuples(
    val_fraction: float,
    test_fraction: float,
    total_size: int,
    starting_test: bool = False,
):
    if starting_test:
        # test => val => train
        test = list(range(0, int(total_size * test_fraction)))
        val = list(range(test[-1] + 1, test[-1] + 1 + int(total_size * val_fraction)))
        train = list(range(val[-1] + 1, total_size))
    else:
        # {test,val}=> train
        test_val_size = int((val_fraction + test_fraction) * total_size)
        train = list(range(test_val_size, total_size))

        if test_val_size == 0:
            test = []
            val = []
            return train, val, test

        # Split the test and validation in chunks.
        chunksize = max(1, min(3, test_val_size // 2))

        nchunks = test_val_size // chunksize

        test = []
        val = []
        s = 0
        for i in range(nchunks):
            if i % 2 == 0:
                val += list(np.arange(s, s + chunksize))
            else:
                test += list(np.arange(s, s + chunksize))
            s += chunksize

        if i % 2 == 0:
            test += list(np.arange(s, test_val_size))
        else:
            p1, p2 = split_in_half(s, test_val_size)
            test += p1
            val += p2

    val, test = adjust_for_imbalance_in_fraction_value(
        val, test, val_fraction, test_fraction, total_size
    )

    return train, val, test



from pathlib import Path

import pytest
from typer.testing import CliRunner

from careamics.cli.main import app

runner = CliRunner()


@pytest.mark.parametrize("algorithm", ["care", "n2n", "n2v"])
def test_conf(tmp_path: Path, algorithm: str):
    config_path = tmp_path / "config.yaml"
    result = runner.invoke(
        app,
        [
            "conf",
            "-d",
            str(tmp_path),
            algorithm,
            "--experiment-name",
            "LevitatingFrog",
            "--axes",
            "YX",
            "--patch-size",
            "64",
            "64",
            "-1",
            "--batch-size",
            "1",
            "--num-epochs",
            "1",
        ],
    )
    assert config_path.is_file()
    assert result.exit_code == 0



"""Module containing convenience function to create `WriteStrategy`."""

from typing import Any, Optional

from careamics.config.support import SupportedData
from careamics.file_io import SupportedWriteType, WriteFunc, get_write_func

from .write_strategy import CacheTiles, WriteImage, WriteStrategy


def create_write_strategy(
    write_type: SupportedWriteType,
    tiled: bool,
    write_func: Optional[WriteFunc] = None,
    write_extension: Optional[str] = None,
    write_func_kwargs: Optional[dict[str, Any]] = None,
) -> WriteStrategy:
    """
    Create a write strategy from convenient parameters.

    Parameters
    ----------
    write_type : {"tiff", "custom"}
        The data type to save as, includes custom.
    tiled : bool
        Whether the prediction will be tiled or not.
    write_func : WriteFunc, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` a function to save the data must be passed. See notes below.
    write_extension : str, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` an extension to save the data with must be passed.
    write_func_kwargs : dict of {str: any}, optional
        Additional keyword arguments to be passed to the save function.

    Returns
    -------
    WriteStrategy
        A strategy for writing predicions.

    Notes
    -----
    The `write_func` function signature must match that of the example below
        ```
        write_func(file_path: Path, img: NDArray, *args, **kwargs) -> None: ...
        ```

    The `write_func_kwargs` will be passed to the `write_func` doing the following:
        ```
        write_func(file_path=file_path, img=img, **kwargs)
        ```
    """
    if write_func_kwargs is None:
        write_func_kwargs = {}

    write_strategy: WriteStrategy
    if not tiled:
        write_func = select_write_func(write_type=write_type, write_func=write_func)
        write_extension = select_write_extension(
            write_type=write_type, write_extension=write_extension
        )
        write_strategy = WriteImage(
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
        )
    else:
        # select CacheTiles or WriteTilesZarr (when implemented)
        write_strategy = _create_tiled_write_strategy(
            write_type=write_type,
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
        )

    return write_strategy


def _create_tiled_write_strategy(
    write_type: SupportedWriteType,
    write_func: Optional[WriteFunc],
    write_extension: Optional[str],
    write_func_kwargs: dict[str, Any],
) -> WriteStrategy:
    """
    Create a tiled write strategy.

    Either `CacheTiles` for caching tiles until a whole image is predicted or
    `WriteTilesZarr` for writing tiles directly to disk.

    Parameters
    ----------
    write_type : {"tiff", "custom"}
        The data type to save as, includes custom.
    write_func : WriteFunc, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` a function to save the data must be passed. See notes below.
    write_extension : str, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` an extension to save the data with must be passed.
    write_func_kwargs : dict of {str: any}
        Additional keyword arguments to be passed to the save function.

    Returns
    -------
    WriteStrategy
        A strategy for writing tiled predictions.

    Raises
    ------
    NotImplementedError
        if `write_type="zarr" is chosen.
    """
    # if write_type == SupportedData.ZARR:
    #    create *args, **kwargs
    #    return WriteTilesZarr(*args, **kwargs)
    # else:
    if write_type == "zarr":
        raise NotImplementedError("Saving to zarr is not implemented yet.")
    else:
        write_func = select_write_func(write_type=write_type, write_func=write_func)
        write_extension = select_write_extension(
            write_type=write_type, write_extension=write_extension
        )
        return CacheTiles(
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
        )


def select_write_func(
    write_type: SupportedWriteType, write_func: Optional[WriteFunc] = None
) -> WriteFunc:
    """
    Return a function to write images.

    If `write_type` is "custom" then `write_func`, otherwise the known write function
    is selected.

    Parameters
    ----------
    write_type : {"tiff", "custom"}
        The data type to save as, includes custom.
    write_func : WriteFunc, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` a function to save the data must be passed. See notes below.

    Returns
    -------
    WriteFunc
        A function for writing images.

    Raises
    ------
    ValueError
        If `write_type="custom"` but `write_func` has not been given.

    Notes
    -----
    The `write_func` function signature must match that of the example below
        ```
        write_func(file_path: Path, img: NDArray, *args, **kwargs) -> None: ...
        ```
    """
    if write_type == SupportedData.CUSTOM:
        if write_func is None:
            raise ValueError(
                "A save function must be provided for custom data types."
                # TODO: link to how save functions should be implemented
            )
        else:
            write_func = write_func
    else:
        write_func = get_write_func(write_type)
    return write_func


def select_write_extension(
    write_type: SupportedWriteType, write_extension: Optional[str] = None
) -> str:
    """
    Return an extension to add to file paths.

    If `write_type` is "custom" then `write_extension`, otherwise the known
    write extension is selected.

    Parameters
    ----------
    write_type : {"tiff", "custom"}
        The data type to save as, includes custom.
    write_extension : str, optional
        If a known `write_type` is selected this argument is ignored. For a custom
        `write_type` an extension to save the data with must be passed.

    Returns
    -------
    str
        The extension to be added to file paths.

    Raises
    ------
    ValueError
        If `self.save_type="custom"` but `save_extension` has not been given.
    """
    write_type_: SupportedData = SupportedData(write_type)  # new variable for mypy
    if write_type_ == SupportedData.CUSTOM:
        if write_extension is None:
            raise ValueError("A save extension must be provided for custom data types.")
        else:
            write_extension = write_extension
    else:
        # kind of a weird pattern -> reason to move get_extension from SupportedData
        write_extension = write_type_.get_extension(write_type_)
    return write_extension



"""Test write strategy factory module."""

from pathlib import Path

import numpy as np
import pytest
from numpy.typing import NDArray

from careamics.file_io.write import write_tiff
from careamics.lightning.callbacks.prediction_writer_callback import (
    CacheTiles,
    WriteImage,
    create_write_strategy,
    select_write_extension,
    select_write_func,
)


def save_numpy(file_path: Path, img: NDArray, *args, **kwargs) -> None:
    """Example custom save function."""
    np.save(file_path, img, *args, **kwargs)


def test_create_write_strategy_tiff_tiled():
    """Test write strategy creation for tiled tiff."""
    write_strategy = create_write_strategy(write_type="tiff", tiled=True)

    assert isinstance(write_strategy, CacheTiles)
    assert write_strategy.write_func is write_tiff
    assert write_strategy.write_extension == ".tiff"
    assert write_strategy.write_func_kwargs == {}


def test_create_write_strategy_tiff_untiled():
    """Test write strategy creation for untiled tiff."""
    write_strategy = create_write_strategy(write_type="tiff", tiled=False)

    assert isinstance(write_strategy, WriteImage)
    assert write_strategy.write_func is write_tiff
    assert write_strategy.write_extension == ".tiff"
    assert write_strategy.write_func_kwargs == {}


def test_create_write_strategy_custom_tiled():
    """Test write strategy creation for tiled custom type."""
    write_strategy = create_write_strategy(
        write_type="custom", tiled=True, write_func=save_numpy, write_extension=".npy"
    )
    assert isinstance(write_strategy, CacheTiles)
    assert write_strategy.write_func is save_numpy
    assert write_strategy.write_extension == ".npy"
    assert write_strategy.write_func_kwargs == {}


def test_create_write_strategy_custom_untiled():
    """Test write strategy creation for untiled custom type."""
    write_strategy = create_write_strategy(
        write_type="custom", tiled=False, write_func=save_numpy, write_extension=".npy"
    )
    assert isinstance(write_strategy, WriteImage)
    assert write_strategy.write_func is save_numpy
    assert write_strategy.write_extension == ".npy"
    assert write_strategy.write_func_kwargs == {}


def test_select_write_func_tiff():
    """Test tiff write function is selected correctly."""
    write_func = select_write_func(write_type="tiff", write_func=None)
    assert write_func is write_tiff


@pytest.mark.parametrize("write_func", [None, save_numpy])
def test_select_write_func_custom(write_func):
    """Test custom write func is returned or raises error if None."""
    if write_func is None:
        with pytest.raises(ValueError):
            write_func = select_write_func("custom", write_func)
        return
    write_func = select_write_func("custom", write_func)
    assert write_func is save_numpy


def test_select_write_extension_tiff():
    """Test tiff extension is selected correctly."""
    write_extension = select_write_extension(write_type="tiff", write_extension=None)
    assert write_extension == ".tiff"


@pytest.mark.parametrize("write_extension", [None, ".npy"])
def test_select_write_extension_custom(write_extension):
    """Test custom extension is returned or raises if None."""
    if write_extension is None:
        with pytest.raises(ValueError):
            write_extension = select_write_extension("custom", write_extension)
        return
    write_extension = select_write_extension("custom", write_extension)
    assert write_extension == ".npy"



from pathlib import Path

import pytest

from careamics.config import (
    Configuration,
    load_configuration,
    save_configuration,
)
from careamics.config.support import (
    SupportedAlgorithm,
    SupportedPixelManipulation,
    SupportedTransform,
)


@pytest.mark.parametrize("name", ["Sn4K3", "C4_M e-L"])
def test_valid_names(minimum_configuration: dict, name: str):
    """Test valid names (letters, numbers, spaces, dashes and underscores)."""
    minimum_configuration["experiment_name"] = name
    myconf = Configuration(**minimum_configuration)
    assert myconf.experiment_name == name


@pytest.mark.parametrize("name", ["", "   ", "#", "/", "^", "%", ",", ".", "a=b"])
def test_invalid_names(minimum_configuration: dict, name: str):
    """Test that invalid names raise an error."""
    minimum_configuration["experiment_name"] = name
    with pytest.raises(ValueError):
        Configuration(**minimum_configuration)


def test_3D_algorithm_and_data_compatibility(minimum_configuration: dict):
    """Test that errors are raised if algorithm `is_3D` and data axes are
    incompatible.
    """
    # 3D but no Z in axes
    minimum_configuration["algorithm_config"]["model"]["conv_dims"] = 3
    config = Configuration(**minimum_configuration)
    assert config.algorithm_config.model.conv_dims == 2

    # 2D but Z in axes
    minimum_configuration["algorithm_config"]["model"]["conv_dims"] = 2
    minimum_configuration["data_config"]["axes"] = "ZYX"
    minimum_configuration["data_config"]["patch_size"] = [64, 64, 64]
    config = Configuration(**minimum_configuration)
    assert config.algorithm_config.model.conv_dims == 3


def test_set_3D(minimum_configuration: dict):
    """Test the set 3D method."""
    conf = Configuration(**minimum_configuration)

    # set to 3D
    conf.set_3D(True, "ZYX", [64, 64, 64])
    assert conf.data_config.axes == "ZYX"
    assert conf.data_config.patch_size == [64, 64, 64]
    assert conf.algorithm_config.model.conv_dims == 3

    # set to 2D
    conf.set_3D(False, "SYX", [64, 64])
    assert conf.data_config.axes == "SYX"
    assert conf.data_config.patch_size == [64, 64]
    assert conf.algorithm_config.model.conv_dims == 2


def test_algorithm_and_data_default_transforms(minimum_configuration: dict):
    """Test that the default data transforms are compatible with n2v."""
    minimum_configuration["algorithm_config"] = {
        "algorithm": "n2v",
        "loss": "n2v",
        "model": {
            "architecture": "UNet",
        },
    }
    Configuration(**minimum_configuration)


@pytest.mark.parametrize(
    "algorithm, strategy",
    [
        ("n2v", SupportedPixelManipulation.UNIFORM.value),
        ("n2v", SupportedPixelManipulation.MEDIAN.value),
        ("n2v2", SupportedPixelManipulation.UNIFORM.value),
        ("n2v2", SupportedPixelManipulation.MEDIAN.value),
    ],
)
def test_n2v2_and_transforms(minimum_configuration: dict, algorithm, strategy):
    """Test that the manipulation strategy is corrected if the data transforms are
    incompatible with n2v2."""
    use_n2v2 = algorithm == "n2v2"
    minimum_configuration["algorithm_config"] = {
        "algorithm": "n2v",
        "loss": "n2v",
        "model": {
            "architecture": "UNet",
            "n2v2": use_n2v2,
        },
    }

    expected_strategy = (
        SupportedPixelManipulation.MEDIAN.value
        if use_n2v2
        else SupportedPixelManipulation.UNIFORM.value
    )

    # missing ManipulateN2V
    minimum_configuration["data_config"]["transforms"] = [
        {"name": SupportedTransform.XY_FLIP.value}
    ]
    config = Configuration(**minimum_configuration)
    assert len(config.data_config.transforms) == 2
    assert (
        config.data_config.transforms[-1].name
        == SupportedTransform.N2V_MANIPULATE.value
    )
    assert config.data_config.transforms[-1].strategy == expected_strategy

    # passing ManipulateN2V with the wrong strategy
    minimum_configuration["data_config"]["transforms"] = [
        {
            "name": SupportedTransform.N2V_MANIPULATE.value,
            "strategy": strategy,
        }
    ]
    config = Configuration(**minimum_configuration)
    assert config.data_config.transforms[-1].strategy == expected_strategy


def test_setting_n2v2(minimum_configuration: dict):
    # make sure we use n2v
    minimum_configuration["algorithm_config"][
        "algorithm"
    ] = SupportedAlgorithm.N2V.value

    # test config
    config = Configuration(**minimum_configuration)
    assert config.algorithm_config.algorithm == SupportedAlgorithm.N2V.value
    assert not config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.UNIFORM.value
    )

    # set N2V2
    config.set_N2V2(True)
    assert config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.MEDIAN.value
    )

    # set back to N2V
    config.set_N2V2(False)
    assert not config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.UNIFORM.value
    )


def test_config_to_yaml(tmp_path: Path, minimum_configuration: dict):
    """Test that we can export a config to yaml and load it back"""

    # test that we can instantiate a config
    myconf = Configuration(**minimum_configuration)

    # export to yaml
    yaml_path = save_configuration(myconf, tmp_path)
    assert yaml_path.exists()

    # load from yaml
    my_other_conf = load_configuration(yaml_path)
    assert my_other_conf == myconf


def test_config_to_yaml_wrong_path(tmp_path: Path, minimum_configuration: dict):
    """Test that an error is raised when the path is not a directory and not a .yml"""

    # test that we can instantiate a config
    myconf = Configuration(**minimum_configuration)

    # export to yaml
    yaml_path = tmp_path / "tmp.txt"
    with pytest.raises(ValueError):
        save_configuration(myconf, yaml_path)

    # existing file
    yaml_path.touch()
    with pytest.raises(ValueError):
        save_configuration(myconf, yaml_path)



import numpy as np
import pytest
import torch

from careamics.utils.metrics import (
    _zero_mean,
    multiscale_ssim,
    psnr,
    scale_invariant_psnr,
)

# TODO: add missing test cases
# - `psnr` result
# - `_fix` and `fix_range`
# - `RunningPSNR`
# - `multiscale_ssim` result


@pytest.mark.parametrize(
    "x",
    [
        np.array([1, 2, 3, 4, 5]),
        np.array([[1, 2, 3], [4, 5, 6]]),
        torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0]),
        torch.tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
    ],
)
def test_zero_mean(x: np.ndarray):
    assert np.allclose(_zero_mean(x), x - x.mean())


# TODO: with 2 identical arrays, shouldn't the result be `inf`?
@pytest.mark.parametrize(
    "gt, pred, result",
    [
        (np.array([1, 2, 3, 4, 5, 6]), np.array([1, 2, 3, 4, 5, 6]), 332.22),
        (np.array([[1, 2, 3], [4, 5, 6]]), np.array([[1, 2, 3], [4, 5, 6]]), 332.22),
    ],
)
def test_scale_invariant_psnr(gt: np.ndarray, pred: np.ndarray, result: float):
    assert scale_invariant_psnr(gt, pred) == pytest.approx(result, rel=5e-3)


@pytest.mark.parametrize(
    "data_type",
    [np.uint8, np.uint16, np.float32, np.int64],
)
def test_psnr(data_type: np.dtype):
    gt_ = np.random.randint(0, 255, size=(16, 16)).astype(data_type)
    pred_ = np.random.randint(0, 255, size=(16, 16)).astype(data_type)

    assert psnr(gt_, pred_, 255) is not None

    range_ = gt_.max() - gt_.min()
    assert psnr(gt_, pred_, range_) is not None
    # add check on the result


@pytest.mark.parametrize("type_", ["torch", "numpy"])
@pytest.mark.parametrize("num_ch", [1, 4])
def test_multiscale_ssim(type_: str, num_ch: int):
    if type_ == "torch":
        gt = torch.rand(4, 256, 256, num_ch)
        pred = torch.rand(4, 256, 256, num_ch)
    elif type_ == "numpy":
        gt = np.random.rand(4, 256, 256, num_ch)
        pred = np.random.rand(4, 256, 256, num_ch)

    rinv_mssim = multiscale_ssim(gt, pred, range_invariant=True)
    mssim = multiscale_ssim(gt, pred, range_invariant=False)
    assert len(rinv_mssim) == num_ch
    assert len(mssim) == num_ch



"""Iterable prediction dataset used to load data file by file."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Generator

from numpy.typing import NDArray
from torch.utils.data import IterableDataset

from careamics.file_io.read import read_tiff
from careamics.transforms import Compose

from ..config import InferenceConfig
from ..config.transformations import NormalizeModel
from .dataset_utils import iterate_over_files


class IterablePredDataset(IterableDataset):
    """Simple iterable prediction dataset.

    Parameters
    ----------
    prediction_config : InferenceConfig
        Inference configuration.
    src_files : List[Path]
        List of data files.
    read_source_func : Callable, optional
        Read source function for custom types, by default read_tiff.
    **kwargs : Any
        Additional keyword arguments, unused.

    Attributes
    ----------
    data_path : Union[str, Path]
        Path to the data, must be a directory.
    axes : str
        Description of axes in format STCZYX.
    mean : Optional[float], optional
        Expected mean of the dataset, by default None.
    std : Optional[float], optional
        Expected standard deviation of the dataset, by default None.
    patch_transform : Optional[Callable], optional
        Patch transform callable, by default None.
    """

    def __init__(
        self,
        prediction_config: InferenceConfig,
        src_files: list[Path],
        read_source_func: Callable = read_tiff,
        **kwargs: Any,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        prediction_config : InferenceConfig
            Inference configuration.
        src_files : list of pathlib.Path
            List of data files.
        read_source_func : Callable, optional
            Read source function for custom types, by default read_tiff.
        **kwargs : Any
            Additional keyword arguments, unused.

        Raises
        ------
        ValueError
            If mean and std are not provided in the inference configuration.
        """
        self.prediction_config = prediction_config
        self.data_files = src_files
        self.axes = prediction_config.axes
        self.read_source_func = read_source_func

        # check mean and std and create normalize transform
        if (
            self.prediction_config.image_means is None
            or self.prediction_config.image_stds is None
        ):
            raise ValueError("Mean and std must be provided for prediction.")
        else:
            self.image_means = self.prediction_config.image_means
            self.image_stds = self.prediction_config.image_stds

        # instantiate normalize transform
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(
                    image_means=self.image_means,
                    image_stds=self.image_stds,
                )
            ],
        )

    def __iter__(
        self,
    ) -> Generator[NDArray, None, None]:
        """
        Iterate over data source and yield single patch.

        Yields
        ------
        NDArray
            Single patch.
        """
        assert (
            self.image_means is not None and self.image_stds is not None
        ), "Mean and std must be provided"

        for sample, _ in iterate_over_files(
            self.prediction_config,
            self.data_files,
            read_source_func=self.read_source_func,
        ):
            # sample has S dimension
            for i in range(sample.shape[0]):

                transformed_sample, _ = self.patch_transform(patch=sample[i])

                yield transformed_sample



"""CAREamics PyTorch Lightning modules."""

__all__ = [
    "FCNModule",
    "VAEModule",
    "create_careamics_module",
    "TrainDataModule",
    "create_train_datamodule",
    "PredictDataModule",
    "create_predict_datamodule",
    "HyperParametersCallback",
    "ProgressBarCallback",
]

from .callbacks import HyperParametersCallback, ProgressBarCallback
from .lightning_module import FCNModule, VAEModule, create_careamics_module
from .predict_data_module import PredictDataModule, create_predict_datamodule
from .train_data_module import TrainDataModule, create_train_datamodule



from pathlib import Path
from typing import Union

import numpy as np
import pytest
import torch

from careamics.config.likelihood_model import (
    GaussianLikelihoodConfig,
    NMLikelihoodConfig,
)
from careamics.config.nm_model import GaussianMixtureNMConfig, MultiChannelNMConfig
from careamics.models.lvae.likelihoods import likelihood_factory
from careamics.models.lvae.noise_models import noise_model_factory

pytestmark = pytest.mark.lvae


# TODO: move it under models/lvae/ ??


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3])
@pytest.mark.parametrize("predict_logvar", [None, "pixelwise"])
@pytest.mark.parametrize("logvar_lowerbound", [None, 0.1])
def test_gaussian_likelihood(
    batch_size: int,
    target_ch: int,
    predict_logvar: Union[str, None],
    logvar_lowerbound: Union[float, None],
) -> None:
    config = GaussianLikelihoodConfig(
        predict_logvar=predict_logvar, logvar_lowerbound=logvar_lowerbound
    )
    likelihood = likelihood_factory(config)

    img_size = 64
    inp_ch = target_ch * (1 + int(predict_logvar is not None))
    reconstruction = torch.rand((batch_size, inp_ch, img_size, img_size))
    target = torch.rand((batch_size, target_ch, img_size, img_size))
    out, data = likelihood(reconstruction, target)

    exp_out_shape = (batch_size, target_ch, img_size, img_size)
    assert out.shape == exp_out_shape
    assert out[0].mean() is not None
    assert data["mean"].shape == exp_out_shape
    if predict_logvar == "pixelwise":
        assert data["logvar"].shape == exp_out_shape
    else:
        assert data["logvar"] is None


@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("target_ch", [1, 3, 5])
def test_noise_model_likelihood(
    tmp_path: Path,
    batch_size: int,
    img_size: int,
    target_ch: int,
    create_dummy_noise_model,
) -> None:
    # Instantiate the noise model
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    noise_model_config = MultiChannelNMConfig(noise_models=[gmm] * target_ch)
    nm = noise_model_factory(noise_model_config)

    # Instantiate the likelihood
    inp_shape = (batch_size, target_ch, img_size, img_size)
    reconstruction = target = torch.rand(inp_shape)
    # NOTE: `input_` is actually the output of LVAE decoder
    data_mean = target.mean(dim=(0, 2, 3), keepdim=True)
    data_std = target.std(dim=(0, 2, 3), keepdim=True)
    config = NMLikelihoodConfig(data_mean=data_mean, data_std=data_std)
    likelihood = likelihood_factory(config, noise_model=nm)

    out, data = likelihood(reconstruction, target)
    exp_out_shape = inp_shape
    assert out.shape == exp_out_shape
    assert out[0].mean() is not None
    assert data["mean"].shape == exp_out_shape
    assert data["logvar"] is None



"""
Layer module.

This submodule contains layers used in the CAREamics models.
"""

from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch.nn import functional as F


class Conv_Block(nn.Module):
    """
    Convolution block used in UNets.

    Convolution block consist of two convolution layers with optional batch norm,
    dropout and with a final activation function.

    The parameters are directly mapped to PyTorch Conv2D and Conv3d parameters, see
    PyTorch torch.nn.Conv2d and torch.nn.Conv3d for more information.

    Parameters
    ----------
    conv_dim : int
        Number of dimension of the convolutions, 2 or 3.
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    intermediate_channel_multiplier : int, optional
        Multiplied for the number of output channels, by default 1.
    stride : int, optional
        Stride of the convolutions, by default 1.
    padding : int, optional
        Padding of the convolutions, by default 1.
    bias : bool, optional
        Bias of the convolutions, by default True.
    groups : int, optional
        Controls the connections between inputs and outputs, by default 1.
    activation : str, optional
        Activation function, by default "ReLU".
    dropout_perc : float, optional
        Dropout percentage, by default 0.
    use_batch_norm : bool, optional
        Use batch norm, by default False.
    """

    def __init__(
        self,
        conv_dim: int,
        in_channels: int,
        out_channels: int,
        intermediate_channel_multiplier: int = 1,
        stride: int = 1,
        padding: int = 1,
        bias: bool = True,
        groups: int = 1,
        activation: str = "ReLU",
        dropout_perc: float = 0,
        use_batch_norm: bool = False,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        conv_dim : int
            Number of dimension of the convolutions, 2 or 3.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        intermediate_channel_multiplier : int, optional
            Multiplied for the number of output channels, by default 1.
        stride : int, optional
            Stride of the convolutions, by default 1.
        padding : int, optional
            Padding of the convolutions, by default 1.
        bias : bool, optional
            Bias of the convolutions, by default True.
        groups : int, optional
            Controls the connections between inputs and outputs, by default 1.
        activation : str, optional
            Activation function, by default "ReLU".
        dropout_perc : float, optional
            Dropout percentage, by default 0.
        use_batch_norm : bool, optional
            Use batch norm, by default False.
        """
        super().__init__()
        self.use_batch_norm = use_batch_norm
        self.conv1 = getattr(nn, f"Conv{conv_dim}d")(
            in_channels,
            out_channels * intermediate_channel_multiplier,
            kernel_size=3,
            stride=stride,
            padding=padding,
            bias=bias,
            groups=groups,
        )

        self.conv2 = getattr(nn, f"Conv{conv_dim}d")(
            out_channels * intermediate_channel_multiplier,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=padding,
            bias=bias,
            groups=groups,
        )

        self.batch_norm1 = getattr(nn, f"BatchNorm{conv_dim}d")(
            out_channels * intermediate_channel_multiplier
        )
        self.batch_norm2 = getattr(nn, f"BatchNorm{conv_dim}d")(out_channels)

        self.dropout = (
            getattr(nn, f"Dropout{conv_dim}d")(dropout_perc)
            if dropout_perc > 0
            else None
        )
        self.activation = (
            getattr(nn, f"{activation}")() if activation is not None else nn.Identity()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        if self.use_batch_norm:
            x = self.conv1(x)
            x = self.batch_norm1(x)
            x = self.activation(x)
            x = self.conv2(x)
            x = self.batch_norm2(x)
            x = self.activation(x)
        else:
            x = self.conv1(x)
            x = self.activation(x)
            x = self.conv2(x)
            x = self.activation(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return x


def _unpack_kernel_size(
    kernel_size: Union[Tuple[int, ...], int], dim: int
) -> Tuple[int, ...]:
    """Unpack kernel_size to a tuple of ints.

    Inspired by Kornia implementation. TODO: link

    Parameters
    ----------
    kernel_size : Union[Tuple[int, ...], int]
        Kernel size.
    dim : int
        Number of dimensions.

    Returns
    -------
    Tuple[int, ...]
        Kernel size tuple.
    """
    if isinstance(kernel_size, int):
        kernel_dims = tuple([kernel_size for _ in range(dim)])
    else:
        kernel_dims = kernel_size
    return kernel_dims


def _compute_zero_padding(
    kernel_size: Union[Tuple[int, ...], int], dim: int
) -> Tuple[int, ...]:
    """Utility function that computes zero padding tuple.

    Parameters
    ----------
    kernel_size : Union[Tuple[int, ...], int]
        Kernel size.
    dim : int
        Number of dimensions.

    Returns
    -------
    Tuple[int, ...]
        Zero padding tuple.
    """
    kernel_dims = _unpack_kernel_size(kernel_size, dim)
    return tuple([(kd - 1) // 2 for kd in kernel_dims])


def get_pascal_kernel_1d(
    kernel_size: int,
    norm: bool = False,
    *,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
) -> torch.Tensor:
    """Generate Yang Hui triangle (Pascal's triangle) for a given number.

    Inspired by Kornia implementation. TODO link

    Parameters
    ----------
    kernel_size : int
        Kernel size.
    norm : bool
        Normalize the kernel, by default False.
    device : Optional[torch.device]
        Device of the tensor, by default None.
    dtype : Optional[torch.dtype]
        Data type of the tensor, by default None.

    Returns
    -------
    torch.Tensor
        Pascal kernel.

    Examples
    --------
    >>> get_pascal_kernel_1d(1)
    tensor([1.])
    >>> get_pascal_kernel_1d(2)
    tensor([1., 1.])
    >>> get_pascal_kernel_1d(3)
    tensor([1., 2., 1.])
    >>> get_pascal_kernel_1d(4)
    tensor([1., 3., 3., 1.])
    >>> get_pascal_kernel_1d(5)
    tensor([1., 4., 6., 4., 1.])
    >>> get_pascal_kernel_1d(6)
    tensor([ 1.,  5., 10., 10.,  5.,  1.])
    """
    pre: List[float] = []
    cur: List[float] = []
    for i in range(kernel_size):
        cur = [1.0] * (i + 1)

        for j in range(1, i // 2 + 1):
            value = pre[j - 1] + pre[j]
            cur[j] = value
            if i != 2 * j:
                cur[-j - 1] = value
        pre = cur

    out = torch.tensor(cur, device=device, dtype=dtype)

    if norm:
        out = out / out.sum()

    return out


def _get_pascal_kernel_nd(
    kernel_size: Union[Tuple[int, int], int],
    norm: bool = True,
    dim: int = 2,
    *,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
) -> torch.Tensor:
    """Generate pascal filter kernel by kernel size.

    If kernel_size is an integer the kernel will be shaped as (kernel_size, kernel_size)
    otherwise the kernel will be shaped as kernel_size

    Inspired by Kornia implementation.

    Parameters
    ----------
    kernel_size : Union[Tuple[int, int], int]
        Kernel size for the pascal kernel.
    norm : bool
        Normalize the kernel, by default True.
    dim : int
        Number of dimensions, by default 2.
    device : Optional[torch.device]
        Device of the tensor, by default None.
    dtype : Optional[torch.dtype]
        Data type of the tensor, by default None.

    Returns
    -------
    torch.Tensor
        Pascal kernel.

    Examples
    --------
    >>> _get_pascal_kernel_nd(1)
    tensor([[1.]])
    >>> _get_pascal_kernel_nd(4)
    tensor([[0.0156, 0.0469, 0.0469, 0.0156],
            [0.0469, 0.1406, 0.1406, 0.0469],
            [0.0469, 0.1406, 0.1406, 0.0469],
            [0.0156, 0.0469, 0.0469, 0.0156]])
    >>> _get_pascal_kernel_nd(4, norm=False)
    tensor([[1., 3., 3., 1.],
            [3., 9., 9., 3.],
            [3., 9., 9., 3.],
            [1., 3., 3., 1.]])
    """
    kernel_dims = _unpack_kernel_size(kernel_size, dim)

    kernel = [
        get_pascal_kernel_1d(kd, device=device, dtype=dtype) for kd in kernel_dims
    ]

    if dim == 2:
        kernel = kernel[0][:, None] * kernel[1][None, :]
    elif dim == 3:
        kernel = (
            kernel[0][:, None, None]
            * kernel[1][None, :, None]
            * kernel[2][None, None, :]
        )
    if norm:
        kernel = kernel / torch.sum(kernel)
    return kernel


def _max_blur_pool_by_kernel2d(
    x: torch.Tensor,
    kernel: torch.Tensor,
    stride: int,
    max_pool_size: int,
    ceil_mode: bool,
) -> torch.Tensor:
    """Compute max_blur_pool by a given :math:`CxC_(out, None)xNxN` kernel.

    Inspired by Kornia implementation.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.
    kernel : torch.Tensor
        Kernel tensor.
    stride : int
        Stride.
    max_pool_size : int
        Maximum pool size.
    ceil_mode : bool
        Ceil mode, by default False. Set to True to match output size of conv2d.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    # compute local maxima
    x = F.max_pool2d(
        x, kernel_size=max_pool_size, padding=0, stride=1, ceil_mode=ceil_mode
    )
    # blur and downsample
    padding = _compute_zero_padding((kernel.shape[-2], kernel.shape[-1]), dim=2)
    return F.conv2d(x, kernel, padding=padding, stride=stride, groups=x.size(1))


def _max_blur_pool_by_kernel3d(
    x: torch.Tensor,
    kernel: torch.Tensor,
    stride: int,
    max_pool_size: int,
    ceil_mode: bool,
) -> torch.Tensor:
    """Compute max_blur_pool by a given :math:`CxC_(out, None)xNxNxN` kernel.

    Inspired by Kornia implementation.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.
    kernel : torch.Tensor
        Kernel tensor.
    stride : int
        Stride.
    max_pool_size : int
        Maximum pool size.
    ceil_mode : bool
        Ceil mode, by default False. Set to True to match output size of conv2d.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    # compute local maxima
    x = F.max_pool3d(
        x, kernel_size=max_pool_size, padding=0, stride=1, ceil_mode=ceil_mode
    )
    # blur and downsample
    padding = _compute_zero_padding(
        (kernel.shape[-3], kernel.shape[-2], kernel.shape[-1]), dim=3
    )
    return F.conv3d(x, kernel, padding=padding, stride=stride, groups=x.size(1))


class MaxBlurPool(nn.Module):
    """Compute pools and blurs and downsample a given feature map.

    Inspired by Kornia MaxBlurPool implementation. Equivalent to
    ```nn.Sequential(nn.MaxPool2d(...), BlurPool2D(...))```

    Parameters
    ----------
    dim : int
        Toggles between 2D and 3D.
    kernel_size : Union[Tuple[int, int], int]
        Kernel size for max pooling.
    stride : int
        Stride for pooling.
    max_pool_size : int
        Max kernel size for max pooling.
    ceil_mode : bool
        Ceil mode, by default False. Set to True to match output size of conv2d.
    """

    def __init__(
        self,
        dim: int,
        kernel_size: Union[Tuple[int, int], int],
        stride: int = 2,
        max_pool_size: int = 2,
        ceil_mode: bool = False,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        dim : int
            Dimension of the convolution.
        kernel_size : Union[Tuple[int, int], int]
            Kernel size for max pooling.
        stride : int, optional
            Stride, by default 2.
        max_pool_size : int, optional
            Maximum pool size, by default 2.
        ceil_mode : bool, optional
            Ceil mode, by default False. Set to True to match output size of conv2d.
        """
        super().__init__()
        self.dim = dim
        self.kernel_size = kernel_size
        self.stride = stride
        self.max_pool_size = max_pool_size
        self.ceil_mode = ceil_mode
        kernel = _get_pascal_kernel_nd(kernel_size, norm=True, dim=self.dim)
        self.register_buffer("kernel", kernel, persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the function.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        kernel = self.kernel.to(dtype=x.dtype)
        num_channels = int(x.size(1))
        if self.dim == 2:
            return _max_blur_pool_by_kernel2d(
                x,
                kernel.repeat((num_channels, 1, 1, 1)),
                self.stride,
                self.max_pool_size,
                self.ceil_mode,
            )
        else:
            return _max_blur_pool_by_kernel3d(
                x,
                kernel.repeat((num_channels, 1, 1, 1, 1)),
                self.stride,
                self.max_pool_size,
                self.ceil_mode,
            )



"""Descriptions of the algorithms used in CAREmics."""

from pydantic import BaseModel

CUSTOM = "Custom"
N2V = "Noise2Void"
N2V2 = "N2V2"
STRUCT_N2V = "StructN2V"
STRUCT_N2V2 = "StructN2V2"
N2N = "Noise2Noise"
CARE = "CARE"


N2V_DESCRIPTION = (
    "Noise2Void is a UNet-based self-supervised algorithm that "
    "uses blind-spot training to denoise images. In short, in every "
    "patches during training, random pixels are selected and their "
    "value replaced by a neighboring pixel value. The network is then "
    "trained to predict the original pixel value. The algorithm "
    "relies on the continuity of the signal (neighboring pixels have "
    "similar values) and the pixel-wise independence of the noise "
    "(the noise in a pixel is not correlated with the noise in "
    "neighboring pixels)."
)


class AlgorithmDescription(BaseModel):
    """Description of an algorithm.

    Attributes
    ----------
    description : str
        Description of the algorithm.
    """

    description: str


class N2VDescription(AlgorithmDescription):
    """Description of Noise2Void.

    Attributes
    ----------
    description : str
        Description of Noise2Void.
    """

    description: str = N2V_DESCRIPTION


class N2V2Description(AlgorithmDescription):
    """Description of N2V2.

    Attributes
    ----------
    description : str
        Description of N2V2.
    """

    description: str = (
        "N2V2 is a variant of Noise2Void. "
        + N2V_DESCRIPTION
        + "\nN2V2 introduces blur-pool layers and removed skip "
        "connections in the UNet architecture to remove checkboard "
        "artefacts, a common artefacts ocurring in Noise2Void."
    )


class StructN2VDescription(AlgorithmDescription):
    """Description of StructN2V.

    Attributes
    ----------
    description : str
        Description of StructN2V.
    """

    description: str = (
        "StructN2V is a variant of Noise2Void. "
        + N2V_DESCRIPTION
        + "\nStructN2V uses a linear mask (horizontal or vertical) to replace "
        "the pixel values of neighbors of the masked pixels by a random "
        "value. Such masking allows removing 1D structured noise from the "
        "the images, the main failure case of the original N2V."
    )


class StructN2V2Description(AlgorithmDescription):
    """Description of StructN2V2.

    Attributes
    ----------
    description : str
        Description of StructN2V2.
    """

    description: str = (
        "StructN2V2 is a a variant of Noise2Void that uses both "
        "structN2V and N2V2. "
        + N2V_DESCRIPTION
        + "\nStructN2V2 uses a linear mask (horizontal or vertical) to replace "
        "the pixel values of neighbors of the masked pixels by a random "
        "value. Such masking allows removing 1D structured noise from the "
        "the images, the main failure case of the original N2V."
        "\nN2V2 introduces blur-pool layers and removed skip connections in "
        "the UNet architecture to remove checkboard artefacts, a common "
        "artefacts ocurring in Noise2Void."
    )


class N2NDescription(AlgorithmDescription):
    """Description of Noise2Noise.

    Attributes
    ----------
    description : str
        Description of Noise2Noise.
    """

    description: str = "Noise2Noise"  # TODO


class CAREDescription(AlgorithmDescription):
    """Description of CARE.

    Attributes
    ----------
    description : str
        Description of CARE.
    """

    description: str = "CARE"  # TODO



# @mock.patch("careamics.utils.wandb.wandb")
# def test_wandb_logger(wandb, tmp_path: Path, minimum_config: dict):
#     config = Configuration(**minimum_config)
#     logger = WandBLogging(
#         experiment_name="test", log_path=tmp_path, config=config, model_to_watch=None
#     )

#     logger.log_metrics({"acc": 0.0})
#     wandb.init().log.assert_called_once_with({"acc": 0.0}, commit=True)


# @mock.patch("careamics.utils.wandb.wandb")
# def test_wandb_logger_engine(wandb, minimum_config: dict):
#     config = Configuration(**minimum_config)
#     config.training.use_wandb = True
#     engine = Engine(config=config)
#     if engine.use_wandb:
#         assert engine.wandb is not None
#         assert wandb.run



import numpy as np
import pytest
import torch

from careamics.transforms import ImageRestorationTTA, XYFlip, XYRandomRotate90


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 1, 8, 8),
        (2, 1, 8, 8),
        (1, 3, 8, 8),
        (2, 3, 8, 8),
        # 3D
        (1, 1, 8, 8, 8),
        (2, 1, 8, 8, 8),
        (1, 3, 8, 8, 8),
        (2, 3, 8, 8, 8),
    ],
)
def test_forward(shape):
    """Test that the transformations are correct."""
    array = np.arange(np.prod(shape)).reshape(shape)
    tensor = torch.Tensor(array)

    tta = ImageRestorationTTA()
    augmented = tta.forward(tensor)
    assert len(augmented) == 8

    # check that the shape is the same
    for aug in augmented:
        assert aug.shape == tensor.shape

    # check that all transformed tensors are unique
    for i, aug1 in enumerate(augmented):
        for j, aug2 in enumerate(augmented):
            if i != j:
                assert not torch.allclose(aug1, aug2)


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 1, 8, 8),
        (1, 3, 8, 8),
        # 3D
        (1, 1, 8, 8, 8),
        (1, 3, 8, 8, 8),
    ],
)
def test_same_transforms(shape):
    """Check that all arrays produced by the rotation and flip
    transforms are generated by the TTA.

    Note that the transforms require no sample dimension.
    """
    array = np.arange(np.prod(shape)).reshape(shape)
    tensor = torch.Tensor(array)

    # apply forward transformation
    tta = ImageRestorationTTA()
    augmented = tta.forward(tensor)

    # transofrms
    rot = XYRandomRotate90(seed=42)
    flip = XYFlip(seed=42)

    for _ in range(100):
        rotated, *_ = rot(tensor)
        flipped, *_ = flip(rotated)

        # check that is in the augmented list
        assert any(torch.allclose(torch.Tensor(flipped), aug) for aug in augmented)


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 1, 8, 8),
        (2, 1, 8, 8),
        (1, 3, 8, 8),
        (2, 3, 8, 8),
        # 3D
        (1, 1, 8, 8, 8),
        (2, 1, 8, 8, 8),
        (1, 3, 8, 8, 8),
        (2, 3, 8, 8, 8),
    ],
)
def test_backward(shape):
    """Test that the backward transformation returns the original tensor."""
    array = np.arange(np.prod(shape)).reshape(shape)
    tensor = torch.Tensor(array)

    # apply forward transformation
    tta = ImageRestorationTTA()
    augmented = tta.forward(tensor)

    # apply backward transformation
    original = tta.backward(augmented)
    assert torch.allclose(tensor, original)



import pytest
import torch

from careamics.config import FCNAlgorithmConfig
from careamics.lightning.lightning_module import (
    FCNModule,
    create_careamics_module,
)

# TODO: rename to test_FCN_lightining_module.py


def test_careamics_module(minimum_algorithm_n2v):
    """Test that the minimum algorithm allows instantiating a the Lightning API
    intermediate layer."""
    algo_config = FCNAlgorithmConfig(**minimum_algorithm_n2v)

    # extract model parameters
    model_parameters = algo_config.model.model_dump(exclude_none=True)

    # instantiate FCNModule
    create_careamics_module(
        algorithm=algo_config.algorithm,
        loss=algo_config.loss,
        architecture=algo_config.model.architecture,
        model_parameters=model_parameters,
        optimizer=algo_config.optimizer.name,
        optimizer_parameters=algo_config.optimizer.parameters,
        lr_scheduler=algo_config.lr_scheduler.name,
        lr_scheduler_parameters=algo_config.lr_scheduler.parameters,
    )


def test_careamics_fcn(minimum_algorithm_n2v):
    """Test that the minimum algorithm allows instantiating a CAREamicsKiln."""
    algo_config = FCNAlgorithmConfig(**minimum_algorithm_n2v)

    # instantiate CAREamicsKiln
    FCNModule(algo_config)


@pytest.mark.parametrize(
    "shape",
    [
        (8, 8),
        (16, 16),
        (32, 32),
    ],
)
def test_fcn_module_unet_2D_depth_2_shape(shape):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 2,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 2,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, 1, *shape))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize(
    "shape",
    [
        (8, 8),
        (16, 16),
        (32, 32),
        (64, 64),
        (128, 128),
        (256, 256),
    ],
)
def test_fcn_module_unet_2D_depth_3_shape(shape):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 2,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 3,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, 1, *shape))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize(
    "shape",
    [
        (8, 32, 16),
        (16, 32, 16),
        (8, 32, 32),
        (32, 64, 64),
    ],
)
def test_fcn_module_unet_depth_2_3D(shape):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 3,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 2,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, 1, *shape))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize(
    "shape",
    [
        (8, 64, 64),
        (16, 64, 64),
        (16, 128, 128),
        (32, 128, 128),
    ],
)
def test_fcn_module_unet_depth_3_3D(shape):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 3,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 3,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, 1, *shape))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize(
    "shape",
    [
        (8, 64, 64),
        (16, 64, 64),
        (16, 128, 128),
        (32, 128, 128),
    ],
)
def test_fcn_module_unet_depth_3_3D_n2v2(shape):
    algo_dict = {
        "algorithm": "n2v",
        "model": {
            "architecture": "UNet",
            "conv_dims": 3,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 3,
            "n2v2": True,
        },
        "loss": "n2v",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, 1, *shape))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize("n_channels", [1, 3, 4])
def test_fcn_module_unet_depth_2_channels_2D(n_channels):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 2,
            "in_channels": n_channels,
            "num_classes": n_channels,
            "depth": 2,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, n_channels, 32, 32))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize(
    "n_channels,independent_channels",
    [(1, False), (1, True), (3, False), (3, True), (4, False), (4, True)],
)
def test_fcn_module_unet_depth_3_channels_2D(n_channels, independent_channels):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 2,
            "in_channels": n_channels,
            "num_classes": n_channels,
            "depth": 3,
            "independent_channels": independent_channels,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, n_channels, 64, 64))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize("n_channels", [1, 3, 4])
def test_fcn_module_unet_depth_2_channels_3D(n_channels):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 3,
            "in_channels": n_channels,
            "num_classes": n_channels,
            "depth": 2,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((2, n_channels, 16, 32, 32))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize("n_channels", [1, 3, 4])
def test_fcn_module_unet_depth_3_channels_3D(n_channels):
    algo_dict = {
        "algorithm": "n2n",
        "model": {
            "architecture": "UNet",
            "conv_dims": 3,
            "in_channels": n_channels,
            "num_classes": n_channels,
            "depth": 3,
        },
        "loss": "mae",
    }
    algo_config = FCNAlgorithmConfig(**algo_dict)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # test forward pass
    x = torch.rand((1, n_channels, 16, 64, 64))
    y: torch.Tensor = model.forward(x)
    assert y.shape == x.shape


@pytest.mark.parametrize("tiled", [False, True])
def test_prediction_callback_during_training(minimum_configuration, tiled):
    import numpy as np
    from pytorch_lightning import Callback, Trainer

    from careamics import CAREamist, Configuration
    from careamics.lightning import PredictDataModule, create_predict_datamodule
    from careamics.prediction_utils import convert_outputs

    config = Configuration(**minimum_configuration)

    class CustomPredictAfterValidationCallback(Callback):
        def __init__(self, pred_datamodule: PredictDataModule):
            self.pred_datamodule = pred_datamodule

            # prepare data and setup
            self.pred_datamodule.prepare_data()
            self.pred_datamodule.setup()
            self.pred_dataloader = pred_datamodule.predict_dataloader()

            self.data = None

        def on_validation_epoch_end(self, trainer: Trainer, pl_module):
            if trainer.sanity_checking:  # optional skip
                return

            # update statistics in the prediction dataset for coherence
            # (they can computed on-line by the training dataset)
            self.pred_datamodule.predict_dataset.image_means = (
                trainer.datamodule.train_dataset.image_stats.means
            )
            self.pred_datamodule.predict_dataset.image_stds = (
                trainer.datamodule.train_dataset.image_stats.stds
            )

            # predict on the dataset
            outputs = []
            for idx, batch in enumerate(self.pred_dataloader):
                batch = pl_module.transfer_batch_to_device(batch, pl_module.device, 0)
                outputs.append(pl_module.predict_step(batch, batch_idx=idx))

            self.data = convert_outputs(outputs, self.pred_datamodule.tiled)

    array = np.arange(64 * 64).reshape((64, 64))
    pred_datamodule = create_predict_datamodule(
        pred_data=array,
        data_type=config.data_config.data_type,
        axes=config.data_config.axes,
        image_means=[11.8],  # random placeholder
        image_stds=[3.14],
        tile_size=(16, 16) if tiled else None,
        tile_overlap=(8, 8) if tiled else None,
        batch_size=2,
    )

    predict_after_val_callback = CustomPredictAfterValidationCallback(
        pred_datamodule=pred_datamodule
    )
    engine = CAREamist(config, callbacks=[predict_after_val_callback])
    engine.train(train_source=array)

    assert not np.allclose(array, predict_after_val_callback.data)



from typing import Any, Optional

from pydantic import BaseModel, ConfigDict

from .types import DataSplitType, DataType, TilingMode


# TODO: check if any bool logic can be removed
class DatasetConfig(BaseModel):
    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    data_type: Optional[DataType]
    """Type of the dataset, should be one of DataType"""

    depth3D: Optional[int] = 1
    """Number of slices in 3D. If data is 2D depth3D is equal to 1"""

    datasplit_type: Optional[DataSplitType] = None
    """Whether to return training, validation or test split, should be one of 
    DataSplitType"""

    num_channels: Optional[int] = 2
    """Number of channels in the input"""

    # TODO: remove ch*_fname parameters, should be parsed automatically from a name list
    ch1_fname: Optional[str] = None
    ch2_fname: Optional[str] = None
    ch_input_fname: Optional[str] = None

    input_is_sum: Optional[bool] = False
    """Whether the input is the sum or average of channels"""

    input_idx: Optional[int] = None
    """Index of the channel where the input is stored in the data"""

    target_idx_list: Optional[list[int]] = None
    """Indices of the channels where the targets are stored in the data"""

    # TODO: where are there used?
    start_alpha: Optional[Any] = None
    end_alpha: Optional[Any] = None

    image_size: tuple  # TODO: revisit, new model_config uses tuple
    """Size of one patch of data"""

    grid_size: Optional[int] = None
    """Frame is divided into square grids of this size. A patch centered on a grid 
    having size `image_size` is returned. Grid size not used in training,
    used only during val / test, grid size controls the overlap of the patches"""

    empty_patch_replacement_enabled: Optional[bool] = False
    """Whether to replace the content of one of the channels
    with background with given probability"""
    empty_patch_replacement_channel_idx: Optional[Any] = None
    empty_patch_replacement_probab: Optional[Any] = None
    empty_patch_max_val_threshold: Optional[Any] = None

    uncorrelated_channels: Optional[bool] = False
    """Replace the content in one of the channels with given probability to make 
    channel content 'uncorrelated'"""
    uncorrelated_channel_probab: Optional[float] = 0.5

    poisson_noise_factor: Optional[float] = -1
    """The added poisson noise factor"""

    synthetic_gaussian_scale: Optional[float] = 0.1

    # TODO: set to True in training code, recheck
    input_has_dependant_noise: Optional[bool] = False

    # TODO: sometimes max_val differs between runs with fixed seeds with noise enabled
    enable_gaussian_noise: Optional[bool] = False
    """Whether to enable gaussian noise"""

    # TODO: is this parameter used?
    allow_generation: bool = False

    # TODO: both used in IndexSwitcher, insure correct passing
    training_validtarget_fraction: Any = None
    deterministic_grid: Any = None

    # TODO: why is this not used?
    enable_rotation_aug: Optional[bool] = False

    max_val: Optional[float] = None
    """Maximum data in the dataset. Is calculated for train split, and should be 
    externally set for val and test splits."""

    overlapping_padding_kwargs: Any = None
    """Parameters for np.pad method"""

    # TODO: remove this parameter, controls debug print
    print_vars: Optional[bool] = False

    # Hard-coded parameters (used to be in the config file)
    normalized_input: bool = True
    """If this is set to true, then one mean and stdev is used
                for both channels. Otherwise, two different mean and stdev are used."""
    use_one_mu_std: Optional[bool] = True

    # TODO: is this parameter used?
    train_aug_rotate: Optional[bool] = False
    enable_random_cropping: Optional[bool] = True

    multiscale_lowres_count: Optional[int] = None
    """Number of LC scales"""

    tiling_mode: Optional[TilingMode] = TilingMode.ShiftBoundary

    target_separate_normalization: Optional[bool] = True

    mode_3D: Optional[bool] = False
    """If training in 3D mode or not"""

    trainig_datausage_fraction: Optional[float] = 1.0

    validtarget_random_fraction: Optional[float] = None

    validation_datausage_fraction: Optional[float] = 1.0

    random_flip_z_3D: Optional[bool] = False

    padding_kwargs: Optional[dict] = None



"""LVAE Pydantic model."""

from typing import Literal

from pydantic import ConfigDict, Field, field_validator, model_validator
from typing_extensions import Self

from .architecture_model import ArchitectureModel


# TODO: it is quite confusing to call this LVAEModel, as it is basically a config
class LVAEModel(ArchitectureModel):
    """LVAE model."""

    model_config = ConfigDict(validate_assignment=True, validate_default=True)

    architecture: Literal["LVAE"]
    input_shape: list[int] = Field(default=(64, 64), validate_default=True)
    """Shape of the input patch (C, Z, Y, X) or (C, Y, X) if the data is 2D."""
    encoder_conv_strides: list = Field(default=[2, 2], validate_default=True)
    # TODO make this per hierarchy step ?
    decoder_conv_strides: list = Field(default=[2, 2], validate_default=True)
    """Dimensions (2D or 3D) of the convolutional layers."""
    multiscale_count: int = Field(default=1)
    # TODO there should be a check for multiscale_count in dataset !!

    # 1 - off, len(z_dims) + 1 # TODO Consider starting from 0
    z_dims: list = Field(default=[128, 128, 128, 128])
    output_channels: int = Field(default=1, ge=1)
    encoder_n_filters: int = Field(default=64, ge=8, le=1024)
    decoder_n_filters: int = Field(default=64, ge=8, le=1024)
    encoder_dropout: float = Field(default=0.1, ge=0.0, le=0.9)
    decoder_dropout: float = Field(default=0.1, ge=0.0, le=0.9)
    nonlinearity: Literal[
        "None", "Sigmoid", "Softmax", "Tanh", "ReLU", "LeakyReLU", "ELU"
    ] = Field(
        default="ELU",
    )

    predict_logvar: Literal[None, "pixelwise"] = None
    analytical_kl: bool = Field(default=False)

    @model_validator(mode="after")
    def validate_conv_strides(self: Self) -> Self:
        """
        Validate the convolutional strides.

        Returns
        -------
        list
            Validated strides.

        Raises
        ------
        ValueError
            If the number of strides is not 2.
        """
        if len(self.encoder_conv_strides) < 2 or len(self.encoder_conv_strides) > 3:
            raise ValueError(
                f"Strides must be 2 or 3 (got {len(self.encoder_conv_strides)})."
            )

        if len(self.decoder_conv_strides) < 2 or len(self.decoder_conv_strides) > 3:
            raise ValueError(
                f"Strides must be 2 or 3 (got {len(self.decoder_conv_strides)})."
            )

        # adding 1 to encoder strides for the number of input channels
        if len(self.input_shape) != len(self.encoder_conv_strides):
            raise ValueError(
                f"Input dimensions must be equal to the number of encoder conv strides"
                f" (got {len(self.input_shape)} and {len(self.encoder_conv_strides)})."
            )

        if len(self.encoder_conv_strides) < len(self.decoder_conv_strides):
            raise ValueError(
                f"Decoder can't be 3D when encoder is 2D (got"
                f" {len(self.encoder_conv_strides)} and"
                f"{len(self.decoder_conv_strides)})."
            )

        if any(s < 1 for s in self.encoder_conv_strides) or any(
            s < 1 for s in self.decoder_conv_strides
        ):
            raise ValueError(
                f"All strides must be greater or equal to 1"
                f"(got {self.encoder_conv_strides} and {self.decoder_conv_strides})."
            )
        # TODO: validate max stride size ?
        return self

    @field_validator("input_shape")
    @classmethod
    def validate_input_shape(cls, input_shape: list) -> list:
        """
        Validate the input shape.

        Parameters
        ----------
        input_shape : list
            Shape of the input patch.

        Returns
        -------
        list
            Validated input shape.

        Raises
        ------
        ValueError
            If the number of dimensions is not 3 or 4.
        """
        if len(input_shape) < 2 or len(input_shape) > 3:
            raise ValueError(
                f"Number of input dimensions must be 2 for 2D data 3 for 3D"
                f"(got {len(input_shape)})."
            )

        if any(s < 1 for s in input_shape):
            raise ValueError(
                f"Input shape must be greater than 1 in all dimensions"
                f"(got {input_shape})."
            )
        return input_shape

    @field_validator("encoder_n_filters")
    @classmethod
    def validate_encoder_even(cls, encoder_n_filters: int) -> int:
        """
        Validate that num_channels_init is even.

        Parameters
        ----------
        encoder_n_filters : int
            Number of channels.

        Returns
        -------
        int
            Validated number of channels.

        Raises
        ------
        ValueError
            If the number of channels is odd.
        """
        # if odd
        if encoder_n_filters % 2 != 0:
            raise ValueError(
                f"Number of channels for the bottom layer must be even"
                f" (got {encoder_n_filters})."
            )

        return encoder_n_filters

    @field_validator("decoder_n_filters")
    @classmethod
    def validate_decoder_even(cls, decoder_n_filters: int) -> int:
        """
        Validate that num_channels_init is even.

        Parameters
        ----------
        decoder_n_filters : int
            Number of channels.

        Returns
        -------
        int
            Validated number of channels.

        Raises
        ------
        ValueError
            If the number of channels is odd.
        """
        # if odd
        if decoder_n_filters % 2 != 0:
            raise ValueError(
                f"Number of channels for the bottom layer must be even"
                f" (got {decoder_n_filters})."
            )

        return decoder_n_filters

    @field_validator("z_dims")
    def validate_z_dims(cls, z_dims: tuple) -> tuple:
        """
        Validate the z_dims.

        Parameters
        ----------
        z_dims : tuple
            Tuple of z dimensions.

        Returns
        -------
        tuple
            Validated z dimensions.

        Raises
        ------
        ValueError
            If the number of z dimensions is not 4.
        """
        if len(z_dims) < 2:
            raise ValueError(
                f"Number of z dimensions must be at least 2 (got {len(z_dims)})."
            )

        return z_dims

    @model_validator(mode="after")
    def validate_multiscale_count(self: Self) -> Self:
        """
        Validate the multiscale count.

        Returns
        -------
        Self
            The validated model.
        """
        if self.multiscale_count < 1 or self.multiscale_count > len(self.z_dims) + 1:
            raise ValueError(
                f"Multiscale count must be 1 for LC off or less or equal to the number"
                f" of Z dims + 1 (got {self.multiscale_count} and {len(self.z_dims)})."
            )
        return self

    def set_3D(self, is_3D: bool) -> None:
        """
        Set 3D model by setting the `conv_dims` parameters.

        Parameters
        ----------
        is_3D : bool
            Whether the algorithm is 3D or not.
        """
        if is_3D:
            self.conv_dims = 3
        else:
            self.conv_dims = 2

    def is_3D(self) -> bool:
        """
        Return whether the model is 3D or not.

        Returns
        -------
        bool
            Whether the model is 3D or not.
        """
        return self.conv_dims == 3



"""Algorithms supported by CAREamics."""

from __future__ import annotations

from careamics.utils import BaseEnum


class SupportedAlgorithm(str, BaseEnum):
    """Algorithms available in CAREamics."""

    N2V = "n2v"
    """Noise2Void algorithm, a self-supervised approach based on blind denoising."""

    CARE = "care"
    """Content-aware image restoration, a supervised algorithm used for a variety
    of tasks."""

    N2N = "n2n"
    """Noise2Noise algorithm, a self-supervised denoising scheme based on comparing
    noisy images of the same sample."""

    MUSPLIT = "musplit"
    """An image splitting approach based on ladder VAE architectures."""

    DENOISPLIT = "denoisplit"
    """An image splitting and denoising approach based on ladder VAE architectures."""

    CUSTOM = "custom"
    """Custom algorithm, used for cases where a custom architecture is provided."""

    # PN2V = "pn2v"
    # HDN = "hdn"
    # SEG = "segmentation"



"""Deep-learning model configurations."""

__all__ = [
    "ArchitectureModel",
    "CustomModel",
    "UNetModel",
    "LVAEModel",
    "clear_custom_models",
    "get_custom_model",
    "register_model",
]

from .architecture_model import ArchitectureModel
from .custom_model import CustomModel
from .lvae_model import LVAEModel
from .register_model import clear_custom_models, get_custom_model, register_model
from .unet_model import UNetModel



from pathlib import Path

import numpy as np
import tifffile
from typer.testing import CliRunner

from careamics import CAREamist
from careamics.cli.main import app
from careamics.config import Configuration, save_configuration
from careamics.config.support import SupportedData

runner = CliRunner()


def test_train(tmp_path: Path, minimum_configuration: dict):

    # create & save config
    config_path = tmp_path / "config.yaml"
    config = Configuration(**minimum_configuration)
    config.data_config.data_type = SupportedData.TIFF.value
    save_configuration(config, config_path)

    # training data
    train_array = np.random.rand(32, 32)
    # save files
    train_file = tmp_path / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    # invoke command
    result = runner.invoke(
        app,
        [
            "train",
            str(config_path),
            "-ts",
            str(train_file),
            "-wd",
            str(tmp_path),
        ],
    )
    assert (tmp_path / "checkpoints").is_dir()
    assert len(list((tmp_path / "checkpoints").glob("*.ckpt"))) > 0
    assert result.exit_code == 0


def test_predict_single_file(tmp_path: Path, minimum_configuration: dict):

    # create & save config
    config_path = tmp_path / "config.yaml"
    config = Configuration(**minimum_configuration)
    config.data_config.data_type = SupportedData.TIFF.value
    save_configuration(config, config_path)

    # dummy data
    train_array = np.random.rand(32, 32)
    # save files
    train_file = tmp_path / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    careamist = CAREamist(config, work_dir=tmp_path)
    careamist.train(train_source=train_file)

    checkpoint_path = next(iter((tmp_path / "checkpoints").glob("*.ckpt")))

    result = runner.invoke(
        app, ["predict", str(checkpoint_path), str(train_file), "-wd", str(tmp_path)]
    )
    assert (tmp_path / "predictions").is_dir()
    assert (tmp_path / "predictions" / "train.tiff").is_file()
    assert result.exit_code == 0


def test_predict_directory(tmp_path: Path, minimum_configuration: dict):

    # create & save config
    config_path = tmp_path / "config.yaml"
    config = Configuration(**minimum_configuration)
    config.data_config.data_type = SupportedData.TIFF.value
    save_configuration(config, config_path)

    n_files = 2
    data_dir = tmp_path / "data"
    data_dir.mkdir()

    # dummy data
    for i in range(n_files):
        train_array = np.random.rand(32, 32)
        # save files
        train_file = data_dir / f"train_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    careamist = CAREamist(config, work_dir=tmp_path)
    careamist.train(train_source=data_dir)

    checkpoint_path = next(iter((tmp_path / "checkpoints").glob("*.ckpt")))

    result = runner.invoke(
        app, ["predict", str(checkpoint_path), str(data_dir), "-wd", str(tmp_path)]
    )
    assert (tmp_path / "predictions").is_dir()
    for i in range(n_files):
        assert (tmp_path / "predictions" / f"train_{i}.tiff").is_file()
    assert result.exit_code == 0



"""Normalization and denormalization transforms for image patches."""

from typing import Optional

import numpy as np
from numpy.typing import NDArray

from careamics.transforms.transform import Transform


def _reshape_stats(stats: list[float], ndim: int) -> NDArray:
    """Reshape stats to match the number of dimensions of the input image.

    This allows to broadcast the stats (mean or std) to the image dimensions, and
    thus directly perform a vectorial calculation.

    Parameters
    ----------
    stats : list of float
        List of stats, mean or standard deviation.
    ndim : int
        Number of dimensions of the image, including the C channel.

    Returns
    -------
    NDArray
        Reshaped stats.
    """
    return np.array(stats)[(..., *[np.newaxis] * (ndim - 1))]


class Normalize(Transform):
    """
    Normalize an image or image patch.

    Normalization is a zero mean and unit variance. This transform expects C(Z)YX
    dimensions.

    Not that an epsilon value of 1e-6 is added to the standard deviation to avoid
    division by zero and that it returns a float32 image.

    Parameters
    ----------
    image_means : list of float
        Mean value per channel.
    image_stds : list of float
        Standard deviation value per channel.
    target_means : list of float, optional
        Target mean value per channel, by default None.
    target_stds : list of float, optional
        Target standard deviation value per channel, by default None.

    Attributes
    ----------
    image_means : list of float
        Mean value per channel.
    image_stds : list of float
        Standard deviation value per channel.
    target_means :list of float, optional
        Target mean value per channel, by default None.
    target_stds : list of float, optional
        Target standard deviation value per channel, by default None.
    """

    def __init__(
        self,
        image_means: list[float],
        image_stds: list[float],
        target_means: Optional[list[float]] = None,
        target_stds: Optional[list[float]] = None,
    ):
        """Constructor.

        Parameters
        ----------
        image_means : list of float
            Mean value per channel.
        image_stds : list of float
            Standard deviation value per channel.
        target_means : list of float, optional
            Target mean value per channel, by default None.
        target_stds : list of float, optional
            Target standard deviation value per channel, by default None.
        """
        self.image_means = image_means
        self.image_stds = image_stds
        self.target_means = target_means
        self.target_stds = target_stds

        self.eps = 1e-6

    def __call__(
        self,
        patch: np.ndarray,
        target: Optional[NDArray] = None,
        **additional_arrays: NDArray,
    ) -> tuple[NDArray, Optional[NDArray], dict[str, NDArray]]:
        """Apply the transform to the source patch and the target (optional).

        Parameters
        ----------
        patch : NDArray
            Patch, 2D or 3D, shape C(Z)YX.
        target : NDArray, optional
            Target for the patch, by default None.
        **additional_arrays : NDArray
            Additional arrays that will be transformed identically to `patch` and
            `target`.

        Returns
        -------
        tuple of NDArray
            Transformed patch and target, the target can be returned as `None`.
        """
        if len(self.image_means) != patch.shape[0]:
            raise ValueError(
                f"Number of means (got a list of size {len(self.image_means)}) and "
                f"number of channels (got shape {patch.shape} for C(Z)YX) do not match."
            )
        if len(additional_arrays) != 0:
            raise NotImplementedError(
                "Transforming additional arrays is currently not supported for "
                "`Normalize`."
            )

        # reshape mean and std and apply the normalization to the patch
        means = _reshape_stats(self.image_means, patch.ndim)
        stds = _reshape_stats(self.image_stds, patch.ndim)
        norm_patch = self._apply(patch, means, stds)

        # same for the target patch
        if (
            target is not None
            and self.target_means is not None
            and self.target_stds is not None
        ):
            target_means = _reshape_stats(self.target_means, target.ndim)
            target_stds = _reshape_stats(self.target_stds, target.ndim)
            norm_target = self._apply(target, target_means, target_stds)
        else:
            norm_target = None

        return norm_patch, norm_target, additional_arrays

    def _apply(self, patch: NDArray, mean: NDArray, std: NDArray) -> NDArray:
        """
        Apply the transform to the image.

        Parameters
        ----------
        patch : NDArray
            Image patch, 2D or 3D, shape C(Z)YX.
        mean : NDArray
            Mean values.
        std : NDArray
            Standard deviations.

        Returns
        -------
        NDArray
            Normalized image patch.
        """
        return ((patch - mean) / (std + self.eps)).astype(np.float32)


class Denormalize:
    """
    Denormalize an image.

    Denormalization is performed expecting a zero mean and unit variance input. This
    transform expects C(Z)YX dimensions.

    Note that an epsilon value of 1e-6 is added to the standard deviation to avoid
    division by zero during the normalization step, which is taken into account during
    denormalization.

    Parameters
    ----------
    image_means : list or tuple of float
        Mean value per channel.
    image_stds : list or tuple of float
        Standard deviation value per channel.

    """

    def __init__(
        self,
        image_means: list[float],
        image_stds: list[float],
    ):
        """Constructor.

        Parameters
        ----------
        image_means : list of float
            Mean value per channel.
        image_stds : list of float
            Standard deviation value per channel.
        """
        self.image_means = image_means
        self.image_stds = image_stds

        self.eps = 1e-6

    def __call__(self, patch: NDArray) -> NDArray:
        """Reverse the normalization operation for a batch of patches.

        Parameters
        ----------
        patch : NDArray
            Patch, 2D or 3D, shape BC(Z)YX.

        Returns
        -------
        NDArray
            Transformed array.
        """
        if len(self.image_means) != patch.shape[1]:
            raise ValueError(
                f"Number of means (got a list of size {len(self.image_means)}) and "
                f"number of channels (got shape {patch.shape} for BC(Z)YX) do not "
                f"match."
            )

        means = _reshape_stats(self.image_means, patch.ndim)
        stds = _reshape_stats(self.image_stds, patch.ndim)

        denorm_array = self._apply(
            patch,
            np.swapaxes(means, 0, 1),  # swap axes as C channel is axis 1
            np.swapaxes(stds, 0, 1),
        )

        return denorm_array.astype(np.float32)

    def _apply(self, array: NDArray, mean: NDArray, std: NDArray) -> NDArray:
        """
        Apply the transform to the image.

        Parameters
        ----------
        array : NDArray
            Image patch, 2D or 3D, shape C(Z)YX.
        mean : NDArray
            Mean values.
        std : NDArray
            Standard deviations.

        Returns
        -------
        NDArray
            Denormalized image array.
        """
        return array * (std + self.eps) + mean



"""Iterable tiled prediction dataset used to load data file by file."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Generator

from numpy.typing import NDArray
from torch.utils.data import IterableDataset

from careamics.file_io.read import read_tiff
from careamics.transforms import Compose

from ..config import InferenceConfig
from ..config.tile_information import TileInformation
from ..config.transformations import NormalizeModel
from .dataset_utils import iterate_over_files
from .tiling import extract_tiles


class IterableTiledPredDataset(IterableDataset):
    """Tiled prediction dataset.

    Parameters
    ----------
    prediction_config : InferenceConfig
        Inference configuration.
    src_files : list of pathlib.Path
        List of data files.
    read_source_func : Callable, optional
        Read source function for custom types, by default read_tiff.
    **kwargs : Any
        Additional keyword arguments, unused.

    Attributes
    ----------
    data_path : str or pathlib.Path
        Path to the data, must be a directory.
    axes : str
        Description of axes in format STCZYX.
    mean : float, optional
        Expected mean of the dataset, by default None.
    std : float, optional
        Expected standard deviation of the dataset, by default None.
    patch_transform : Callable, optional
        Patch transform callable, by default None.
    """

    def __init__(
        self,
        prediction_config: InferenceConfig,
        src_files: list[Path],
        read_source_func: Callable = read_tiff,
        **kwargs: Any,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        prediction_config : InferenceConfig
            Inference configuration.
        src_files : List[Path]
            List of data files.
        read_source_func : Callable, optional
            Read source function for custom types, by default read_tiff.
        **kwargs : Any
            Additional keyword arguments, unused.

        Raises
        ------
        ValueError
            If mean and std are not provided in the inference configuration.
        """
        if (
            prediction_config.tile_size is None
            or prediction_config.tile_overlap is None
        ):
            raise ValueError(
                "Tile size and overlap must be provided for tiled prediction."
            )

        self.prediction_config = prediction_config
        self.data_files = src_files
        self.axes = prediction_config.axes
        self.tile_size = prediction_config.tile_size
        self.tile_overlap = prediction_config.tile_overlap
        self.read_source_func = read_source_func

        # check mean and std and create normalize transform
        if (
            self.prediction_config.image_means is None
            or self.prediction_config.image_stds is None
        ):
            raise ValueError("Mean and std must be provided for prediction.")
        else:
            self.image_means = self.prediction_config.image_means
            self.image_stds = self.prediction_config.image_stds

            # instantiate normalize transform
            self.patch_transform = Compose(
                transform_list=[
                    NormalizeModel(
                        image_means=self.image_means,
                        image_stds=self.image_stds,
                    )
                ],
            )

    def __iter__(
        self,
    ) -> Generator[tuple[NDArray, TileInformation], None, None]:
        """
        Iterate over data source and yield single patch.

        Yields
        ------
        Generator of NDArray and TileInformation tuple
            Generator of single tiles.
        """
        assert (
            self.image_means is not None and self.image_stds is not None
        ), "Mean and std must be provided"

        for sample, _ in iterate_over_files(
            self.prediction_config,
            self.data_files,
            read_source_func=self.read_source_func,
        ):
            # generate patches, return a generator of single tiles
            patch_gen = extract_tiles(
                arr=sample,
                tile_size=self.tile_size,
                overlaps=self.tile_overlap,
            )

            # apply transform to patches
            for patch_array, tile_info in patch_gen:
                transformed_patch, _ = self.patch_transform(patch=patch_array)

                yield transformed_patch, tile_info



from careamics.utils.base_enum import BaseEnum


class MyEnum(str, BaseEnum):
    A = "a"
    B = "b"
    C = "c"


def test_base_enum():
    """Test that BaseEnum allows the `in` operator with values."""
    assert "b" in MyEnum



"""Module containing different strategies for writing predictions."""

from pathlib import Path
from typing import Any, Optional, Protocol, Sequence, Union

import numpy as np
from numpy.typing import NDArray
from pytorch_lightning import LightningModule, Trainer
from torch.utils.data import DataLoader

from careamics.config.tile_information import TileInformation
from careamics.dataset import IterablePredDataset, IterableTiledPredDataset
from careamics.file_io import WriteFunc
from careamics.prediction_utils import stitch_prediction_single

from .file_path_utils import create_write_file_path, get_sample_file_path


class WriteStrategy(Protocol):
    """Protocol for write strategy classes."""

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: Any,  # TODO: change to expected type
        batch_indices: Optional[Sequence[int]],
        batch: Any,  # TODO: change to expected type
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        WriteStrategy subclasses must contain this function to write a batch.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.
        """


class CacheTiles(WriteStrategy):
    """
    A write strategy that will cache tiles.

    Tiles are cached until a whole image is predicted on. Then the stitched
    prediction is saved.

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    tile_cache : list of numpy.ndarray
        Tiles cached for stitching prediction.
    tile_info_cache : list of TileInformation
        Cached tile information for stitching prediction.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A write strategy that will cache tiles.

        Tiles are cached until a whole image is predicted on. Then the stitched
        prediction is saved.

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

        # where tiles will be cached until a whole image has been predicted
        self.tile_cache: list[NDArray] = []
        self.tile_info_cache: list[TileInformation] = []

    @property
    def last_tiles(self) -> list[bool]:
        """
        List of bool to determine whether each tile in the cache is the last tile.

        Returns
        -------
        list of bool
            Whether each tile in the tile cache is the last tile.
        """
        return [tile_info.last_tile for tile_info in self.tile_info_cache]

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: tuple[NDArray, list[TileInformation]],
        batch_indices: Optional[Sequence[int]],
        batch: tuple[NDArray, list[TileInformation]],
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Cache tiles until the last tile is predicted; save the stitched prediction.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.
        """
        dataloaders: Union[DataLoader, list[DataLoader]] = trainer.predict_dataloaders
        dataloader: DataLoader = (
            dataloaders[dataloader_idx]
            if isinstance(dataloaders, list)
            else dataloaders
        )
        dataset: IterableTiledPredDataset = dataloader.dataset
        if not isinstance(dataset, IterableTiledPredDataset):
            raise TypeError("Prediction dataset is not `IterableTiledPredDataset`.")

        # cache tiles (batches are split into single samples)
        self.tile_cache.extend(np.split(prediction[0], prediction[0].shape[0]))
        self.tile_info_cache.extend(prediction[1])

        # save stitched prediction
        if self._has_last_tile():

            # get image tiles and remove them from the cache
            tiles, tile_infos = self._get_image_tiles()
            self._clear_cache()

            # stitch prediction
            prediction_image = stitch_prediction_single(
                tiles=tiles, tile_infos=tile_infos
            )

            # write prediction
            sample_id = tile_infos[0].sample_id  # need this to select correct file name
            input_file_path = get_sample_file_path(dataset=dataset, sample_id=sample_id)
            file_path = create_write_file_path(
                dirpath=dirpath,
                file_path=input_file_path,
                write_extension=self.write_extension,
            )
            self.write_func(
                file_path=file_path, img=prediction_image[0], **self.write_func_kwargs
            )

    def _has_last_tile(self) -> bool:
        """
        Whether a last tile is contained in the cached tiles.

        Returns
        -------
        bool
            Whether a last tile is contained in the cached tiles.
        """
        return any(self.last_tiles)

    def _clear_cache(self) -> None:
        """Remove the tiles in the cache up to the first last tile."""
        index = self._last_tile_index()
        self.tile_cache = self.tile_cache[index + 1 :]
        self.tile_info_cache = self.tile_info_cache[index + 1 :]

    def _last_tile_index(self) -> int:
        """
        Find the index of the last tile in the tile cache.

        Returns
        -------
        int
            Index of last tile.

        Raises
        ------
        ValueError
            If there is no last tile in the tile cache.
        """
        last_tiles = self.last_tiles
        if not any(last_tiles):
            raise ValueError("No last tile in the tile cache.")
        index = np.where(last_tiles)[0][0]
        return index

    def _get_image_tiles(self) -> tuple[list[NDArray], list[TileInformation]]:
        """
        Get the tiles corresponding to a single image.

        Returns
        -------
        tuple of (list of numpy.ndarray, list of TileInformation)
            Tiles and tile information to stitch together a full image.
        """
        index = self._last_tile_index()
        tiles = self.tile_cache[: index + 1]
        tile_infos = self.tile_info_cache[: index + 1]
        return tiles, tile_infos


class WriteTilesZarr(WriteStrategy):
    """Strategy to write tiles to Zarr file."""

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: Any,
        batch_indices: Optional[Sequence[int]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Write tiles to zarr file.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError


class WriteImage(WriteStrategy):
    """
    A strategy for writing image predictions (i.e. un-tiled predictions).

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A strategy for writing image predictions (i.e. un-tiled predictions).

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

    def write_batch(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: NDArray,
        batch_indices: Optional[Sequence[int]],
        batch: NDArray,
        batch_idx: int,
        dataloader_idx: int,
        dirpath: Path,
    ) -> None:
        """
        Save full images.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning Trainer.
        pl_module : LightningModule
            PyTorch Lightning LightningModule.
        prediction : Any
            Predictions on `batch`.
        batch_indices : sequence of int
            Indices identifying the samples in the batch.
        batch : Any
            Input batch.
        batch_idx : int
            Batch index.
        dataloader_idx : int
            Dataloader index.
        dirpath : Path
            Path to directory to save predictions to.

        Raises
        ------
        TypeError
            If trainer prediction dataset is not `IterablePredDataset`.
        """
        dls: Union[DataLoader, list[DataLoader]] = trainer.predict_dataloaders
        dl: DataLoader = dls[dataloader_idx] if isinstance(dls, list) else dls
        ds: IterablePredDataset = dl.dataset
        if not isinstance(ds, IterablePredDataset):
            raise TypeError("Prediction dataset is not `IterablePredDataset`.")

        for i in range(prediction.shape[0]):
            prediction_image = prediction[0]
            sample_id = batch_idx * dl.batch_size + i
            input_file_path = get_sample_file_path(dataset=ds, sample_id=sample_id)
            file_path = create_write_file_path(
                dirpath=dirpath,
                file_path=input_file_path,
                write_extension=self.write_extension,
            )
            self.write_func(
                file_path=file_path, img=prediction_image, **self.write_func_kwargs
            )



"""Module contiaing tiling manager class."""

# # TODO: remove this file, left as a reference for now.

# from typing import Any, Optional

# import numpy as np
# from numpy.typing import NDArray

# from careamics.config.tile_information import TileInformation
# from careamics.config.validators import check_axes_validity


# def calculate_padding(
#     patch_start_location: NDArray,
#     patch_size: NDArray,
#     data_shape: NDArray,
# ) -> NDArray:
#     patch_end_location = patch_start_location + patch_size

#     pad_before = np.zeros_like(patch_start_location)
#     start_out_of_bounds = patch_start_location < 0
#     pad_before[start_out_of_bounds] = -patch_start_location[start_out_of_bounds]

#     pad_after = np.zeros_like(patch_start_location)
#     end_out_of_bounds = patch_end_location > data_shape
#     pad_after[end_out_of_bounds] = (
#         patch_end_location - data_shape
#     )[end_out_of_bounds]

#     return np.stack([pad_before, pad_after], axis=1)


# def extract_tile(
#     img: np.ndarray,
#     grid_start_loc: tuple[int, ...],
#     patch_size: tuple[int, ...],
#     overlap: tuple[int, ...],
#     padding: bool,
#     padding_kwargs: Optional[dict[str, Any]] = None,
# ) -> NDArray:
#     if padding_kwargs is None:
#         padding_kwargs = {}

#     data_shape = img.shape
#     patch_start_loc = np.array(grid_start_loc) - np.array(overlap) // 2
#     crop_slices = tuple(
#         slice(max(0, start), min(start + size, dim_shape))
#         for start, size, dim_shape in zip(patch_start_loc, patch_size, data_shape)
#     )
#     crop = img[crop_slices]
#     if padding:
#         pad = calculate_padding(
#             patch_start_location=patch_start_loc,
#             patch_size=patch_size,
#             data_shape=data_shape,
#         )
#         crop = np.pad(crop, pad, **padding_kwargs)

#     return crop


# class TilingManager:

#     def __init__(
#         self,
#         data_shape: tuple[int, ...],
#         tile_size: tuple[int, ...],
#         overlaps: tuple[int, ...],
#         trim_boundary: tuple[int, ...],
#     ):
#         # --- validation
#         if len(data_shape) != len(tile_size):
#             raise ValueError(
#                 f"Data shape:{data_shape} and tile size:{tile_size} must have the "
#                 "same dimension"
#             )
#         if len(data_shape) != len(overlaps):
#             raise ValueError(
#                 f"Data shape:{data_shape} and tile overlaps:{overlaps} must have the "
#                 "same dimension"
#             )
#         # overlaps = np.array(tile_size) - np.array(grid_shape)
#         if (np.array(overlaps) < 0).any():
#             raise ValueError(
#                 "Tile overlap must be positive or zero in all dimension."
#             )
#         if ((np.array(overlaps) % 2) != 0).any():
#             # TODO: currently not required by CAREamics tiling,
#             #   -> because floor divide is used.
#             raise ValueError("Tile overlaps must be even.")

#         # initialize attributes
#         self.data_shape = data_shape
#         self.overlaps = overlaps
#         self.grid_shape = tuple(np.array(tile_size) - np.array(overlaps))
#         self.patch_shape = tile_size
#         self.trim_boundary = trim_boundary

#     def compute_tile_info(self, index: int, axes: str):

#         # TODO: better axis validation, data should already be in the form SC(Z)YX

#         # validate axes
#         check_axes_validity(axes)
#         # z will be -1 if not present
#         spatial_axes = [axes.find("Z"), axes.find("Y"), axes.find("X")]

#         # convert to numpy for convenience
#         data_shape = np.array(self.data_shape)
#         patch_shape = np.array(self.patch_shape)

#         # --- calculate stitch coords
#         stitch_coords_start = np.array(self.get_location_from_dataset_idx(index))
#         stitch_coords_end = stitch_coords_start + np.array(self.grid_shape)

#         # --- patch coords
#         patch_coords_start = stitch_coords_start - np.array(self.overlaps) // 2
#         patch_coords_end = patch_coords_start + patch_shape

#         # --- replace out of bounds indices

#         out_of_lower_bound = stitch_coords_start < 0
#         out_of_upper_bound = stitch_coords_end > data_shape

#         stitch_coords_start[out_of_lower_bound] = 0
#         stitch_coords_end[out_of_upper_bound] = data_shape[out_of_upper_bound]

#         # --- calculate overlap crop coords
#         overlap_crop_coords_start = stitch_coords_start - patch_coords_start
#         overlap_crop_coords_end = overlap_crop_coords_start + (
#             stitch_coords_end - stitch_coords_start
#         )

#         # --- combine start and end
#         stitch_coords = tuple(
#             (stitch_coords_start[axis], stitch_coords_end[axis])
#             for axis in spatial_axes
#             if axis != -1
#         )
#         overlap_crop_coords = tuple(
#             (overlap_crop_coords_start[axis], overlap_crop_coords_end[axis])
#             for axis in spatial_axes
#             if axis != -1
#         )

#         channel_axis = axes.find("C")
#         array_shape_processed = tuple(
#             data_shape[axis] for axis in [channel_axis, *spatial_axes] if axis != -1
#         )

#         tile_info = TileInformation(
#             array_shape=array_shape_processed,
#             last_tile=index == self.total_grid_count() - 1,
#             overlap_crop_coords=overlap_crop_coords,
#             stitch_coords=stitch_coords,
#             sample_id=0, # TODO: in iterable dataset this is also always 0 pretty sure
#         )
#         return tile_info

#     def patch_offset(self):
#         return (np.array(self.patch_shape) - np.array(self.grid_shape)) // 2

#     def get_individual_dim_grid_count(self, dim: int):
#         """
#         Returns the number of the grid in the specified dimension, ignoring all other
#         dimensions.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"

#         if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
#             return self.data_shape[dim]
#         elif self.trim_boundary is False:
#             return int(np.ceil(self.data_shape[dim] / self.grid_shape[dim]))
#         else:
#             excess_size = self.patch_shape[dim] - self.grid_shape[dim]
#             return int(
#                 np.floor((self.data_shape[dim] - excess_size) / self.grid_shape[dim])
#             )

#     def total_grid_count(self):
#         """
#         Returns the total number of grids in the dataset.
#         """
#         return self.grid_count(0) * self.get_individual_dim_grid_count(0)

#     def grid_count(self, dim: int):
#         """
#         Returns the total number of grids for one value in the specified dimension.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"
#         if dim == len(self.data_shape) - 1:
#             return 1

#         return self.get_individual_dim_grid_count(dim + 1) * self.grid_count(dim + 1)

#     def get_grid_index(self, dim: int, coordinate: int):
#         """
#         Returns the index of the grid in the specified dimension.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"
#         assert (
#             coordinate < self.data_shape[dim]
#         ), (
#             f"Coordinate {coordinate} is out of bounds for data "
#             f"shape {self.data_shape}"
#         )
#         if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
#             return coordinate
#         elif self.trim_boundary is False:
#             return np.floor(coordinate / self.grid_shape[dim])
#         else:
#             excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
#             # can be <0 if coordinate is in [0,grid_shape[dim]]
#             return max(0, np.floor((coordinate - excess_size) / self.grid_shape[dim]))

#     def dataset_idx_from_grid_idx(self, grid_idx: tuple):
#         """
#         Returns the index of the grid in the dataset.
#         """
#         assert len(grid_idx) == len(
#             self.data_shape
#         ), (
#             f"Dimension indices {grid_idx} must have the same dimension as data "
#             f"shape {self.data_shape}"
#         )
#         index = 0
#         for dim in range(len(grid_idx)):
#             index += grid_idx[dim] * self.grid_count(dim)
#         return index

#     def get_patch_location_from_dataset_idx(self, dataset_idx: int):
#         """
#         Returns the patch location of the grid in the dataset.
#         """
#         location = self.get_location_from_dataset_idx(dataset_idx)
#         offset = self.patch_offset()
#         return tuple(np.array(location) - np.array(offset))

#     def get_dataset_idx_from_grid_location(self, location: tuple):
#         assert len(location) == len(
#             self.data_shape
#         ), (
#             f"Location {location} must have the same dimension as data shape "
#             f"{self.data_shape}"
#         )
#         grid_idx = [
#             self.get_grid_index(dim, location[dim]) for dim in range(len(location))
#         ]
#         return self.dataset_idx_from_grid_idx(tuple(grid_idx))

#     def get_gridstart_location_from_dim_index(self, dim: int, dim_index: int):
#         """
#         Returns the grid-start coordinate of the grid in the specified dimension.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"
#         assert dim_index < self.get_individual_dim_grid_count(
#             dim
#         ), (
#             f"Dimension index {dim_index} is out of bounds for data shape "
#             f"{self.data_shape}"
#         )

#         if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
#             return dim_index
#         elif self.trim_boundary is False:
#             return dim_index * self.grid_shape[dim]
#         else:
#             excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
#             return dim_index * self.grid_shape[dim] + excess_size

#     def get_location_from_dataset_idx(self, dataset_idx: int):
#         grid_idx = []
#         for dim in range(len(self.data_shape)):
#             grid_idx.append(dataset_idx // self.grid_count(dim))
#             dataset_idx = dataset_idx % self.grid_count(dim)
#         location = [
#             self.get_gridstart_location_from_dim_index(dim, grid_idx[dim])
#             for dim in range(len(self.data_shape))
#         ]
#         return tuple(location)

#     def on_boundary(self, dataset_idx: int, dim: int):
#         """
#         Returns True if the grid is on the boundary in the specified dimension.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"

#         if dim > 0:
#             dataset_idx = dataset_idx % self.grid_count(dim - 1)

#         dim_index = dataset_idx // self.grid_count(dim)
#         return (
#             dim_index == 0 or dim_index == self.get_individual_dim_grid_count(dim) - 1
#         )

#     def next_grid_along_dim(self, dataset_idx: int, dim: int):
#         """
#         Returns the index of the grid in the specified dimension in the specified "
#         "direction.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"
#         new_idx = dataset_idx + self.grid_count(dim)
#         if new_idx >= self.total_grid_count():
#             return None
#         return new_idx

#     def prev_grid_along_dim(self, dataset_idx: int, dim: int):
#         """
#         Returns the index of the grid in the specified dimension in the specified "
#         "direction.
#         """
#         assert dim < len(
#             self.data_shape
#         ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
#         assert dim >= 0, "Dimension must be greater than or equal to 0"
#         new_idx = dataset_idx - self.grid_count(dim)
#         if new_idx < 0:
#             return None


# if __name__ == "__main__":
#     data_shape = (1, 1, 103, 103, 2)
#     grid_shape = (1, 1, 16, 16, 2)
#     patch_shape = (1, 1, 32, 32, 2)
#     overlap = tuple(np.array(patch_shape) - np.array(grid_shape))

#     trim_boundary = False
#     manager = TilingManager(
#         data_shape=data_shape,
#         tile_size=patch_shape,
#         overlaps=overlap,
#         trim_boundary=trim_boundary,
#     )
#     gc = manager.total_grid_count()
#     print("Grid count", gc)
#     for i in range(gc):
#         loc = manager.get_location_from_dataset_idx(i)
#         print(i, loc)
#         inferred_i = manager.get_dataset_idx_from_grid_location(loc)
#         assert i == inferred_i, f"Index mismatch: {i} != {inferred_i}"

#     for i in range(5):
#         print(manager.on_boundary(40, i))



"""Functions used to create a README.md file for BMZ export."""

from pathlib import Path

import yaml

from careamics.config import Configuration
from careamics.utils import cwd, get_careamics_home


def _yaml_block(yaml_str: str) -> str:
    """Return a markdown code block with a yaml string.

    Parameters
    ----------
    yaml_str : str
        YAML string.

    Returns
    -------
    str
        Markdown code block with the YAML string.
    """
    return f"```yaml\n{yaml_str}\n```"


def readme_factory(
    config: Configuration,
    careamics_version: str,
    data_description: str,
) -> Path:
    """Create a README file for the model.

    `data_description` can be used to add more information about the content of the
    data the model was trained on.

    Parameters
    ----------
    config : Configuration
        CAREamics configuration.
    careamics_version : str
        CAREamics version.
    data_description : str
        Description of the data.

    Returns
    -------
    Path
        Path to the README file.
    """
    # create file
    # TODO use tempfile as in the bmz_io module
    with cwd(get_careamics_home()):
        readme = Path("README.md")
        readme.touch()

        # algorithm pretty name
        algorithm_flavour = config.get_algorithm_flavour()
        algorithm_pretty_name = algorithm_flavour + " - CAREamics"

        description = [f"# {algorithm_pretty_name}\n\n"]

        # data description
        description.append("## Data description\n\n")
        description.append(data_description)
        description.append("\n\n")

        # algorithm description
        description.append("## Algorithm description:\n\n")
        description.append(config.get_algorithm_description())
        description.append("\n\n")

        # configuration description
        description.append("## Configuration\n\n")

        description.append(
            f"{algorithm_flavour} was trained using CAREamics (version "
            f"{careamics_version}) using the following configuration:\n\n"
        )

        description.append(_yaml_block(yaml.dump(config.model_dump(exclude_none=True))))
        description.append("\n\n")

        # validation
        description.append("# Validation\n\n")

        description.append(
            "In order to validate the model, we encourage users to acquire a "
            "test dataset with ground-truth data. Comparing the ground-truth data "
            "with the prediction allows unbiased evaluation of the model performances. "
            "This can be done for instance by using metrics such as PSNR, SSIM, or"
            "MicroSSIM. In the absence of ground-truth, inspecting the residual image "
            "(difference between input and predicted image) can be helpful to identify "
            "whether real signal is removed from the input image.\n\n"
        )

        # references
        reference = config.get_algorithm_references()
        if reference != "":
            description.append("## References\n\n")
            description.append(reference)
            description.append("\n\n")

        # links
        description.append(
            "# Links\n\n"
            "- [CAREamics repository](https://github.com/CAREamics/careamics)\n"
            "- [CAREamics documentation](https://careamics.github.io/)\n"
        )

        readme.write_text("".join(description))

        return readme.absolute()



"""Functions relating to reading image files of different formats."""

__all__ = [
    "get_read_func",
    "read_tiff",
    "read_zarr",
    "ReadFunc",
]

from .get_func import ReadFunc, get_read_func
from .tiff import read_tiff
from .zarr import read_zarr



"""A base class for Enum that allows checking if a value is in the Enum."""

from enum import Enum, EnumMeta
from typing import Any


class _ContainerEnum(EnumMeta):
    """Metaclass for Enum with __contains__ method."""

    def __contains__(cls, item: Any) -> bool:
        """Check if an item is in the Enum.

        Parameters
        ----------
        item : Any
            Item to check.

        Returns
        -------
        bool
            True if the item is in the Enum, False otherwise.
        """
        try:
            cls(item)
        except ValueError:
            return False
        return True

    @classmethod
    def has_value(cls, value: Any) -> bool:
        """Check if a value is in the Enum.

        Parameters
        ----------
        value : Any
            Value to check.

        Returns
        -------
        bool
            True if the value is in the Enum, False otherwise.
        """
        return value in cls._value2member_map_


class BaseEnum(Enum, metaclass=_ContainerEnum):
    """Base Enum class, allowing checking if a value is in the enum.

    Example
    -------
    >>> from careamics.utils.base_enum import BaseEnum
    >>> # Define a new enum
    >>> class BaseEnumExtension(BaseEnum):
    ...     VALUE = "value"
    >>> # Check if value is in the enum
    >>> "value" in BaseEnumExtension
    True
    """

    pass



import numpy as np
import pytest

from careamics.config.tile_information import TileInformation


def test_defaults():
    """Test instantiating time information with defaults."""
    tile_info = TileInformation(
        array_shape=np.zeros((1, 6, 6)).shape,
        overlap_crop_coords=((1, 2),),
        stitch_coords=((3, 4),),
        sample_id=0,
    )

    assert tile_info.array_shape == (1, 6, 6)
    assert not tile_info.last_tile


def test_error_on_coords():
    """Test than an error is raised if no coordinates are given."""
    with pytest.raises(ValueError):
        TileInformation(array_shape=(1, 6, 6))


@pytest.mark.parametrize("array_shape", [(6,), (6, 6), (1, 1, 1, 6, 6)])
def test_error_n_dims(array_shape):
    """
    Test that an error is raised if the array shape does not have 3 or 4 dimensions.
    """
    with pytest.raises(ValueError):
        TileInformation(
            array_shape=array_shape,
            overlap_crop_coords=((1, 2),),
            stitch_coords=((3, 4),),
            sample_id=0,
        )


def test_tile_equality():
    """Test whether two tile information objects are equal."""
    t1 = TileInformation(
        array_shape=(1, 6, 6),
        last_tile=True,
        overlap_crop_coords=((1, 2),),
        stitch_coords=((3, 4),),
        sample_id=0,
    )
    t2 = TileInformation(
        array_shape=(1, 6, 6),
        last_tile=True,
        overlap_crop_coords=((1, 2),),
        stitch_coords=((3, 4),),
        sample_id=0,
    )
    assert t1 == t2

    # inequality
    t2.array_shape = (7, 7)
    assert t1 != t2

    t2.array_shape = (6, 6)
    t2.last_tile = False
    assert t1 != t2

    t2.last_tile = True
    t2.overlap_crop_coords = ((2, 3),)
    assert t1 != t2

    t2.overlap_crop_coords = ((1, 2),)
    t2.stitch_coords = ((4, 5),)
    assert t1 != t2



"""Utility function to get RAM size."""

import psutil


def get_ram_size() -> int:
    """
    Get RAM size in mbytes.

    Returns
    -------
    int
        RAM size in mbytes.
    """
    return psutil.virtual_memory().available / 1024**2



"""Pydantic model representing CAREamics prediction configuration."""

from __future__ import annotations

from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from typing_extensions import Self

from .validators import check_axes_validity, patch_size_ge_than_8_power_of_2


class InferenceConfig(BaseModel):
    """Configuration class for the prediction model."""

    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    data_type: Literal["array", "tiff", "custom"]  # As defined in SupportedData
    """Type of input data: numpy.ndarray (array) or path (tiff or custom)."""

    tile_size: Optional[Union[list[int]]] = Field(
        default=None, min_length=2, max_length=3
    )
    """Tile size of prediction, only effective if `tile_overlap` is specified."""

    tile_overlap: Optional[Union[list[int]]] = Field(
        default=None, min_length=2, max_length=3
    )
    """Overlap between tiles, only effective if `tile_size` is specified."""

    axes: str
    """Data axes (TSCZYX) in the order of the input data."""

    image_means: list = Field(..., min_length=0, max_length=32)
    """Mean values for each input channel."""

    image_stds: list = Field(..., min_length=0, max_length=32)
    """Standard deviation values for each input channel."""

    # TODO only default TTAs are supported for now
    tta_transforms: bool = Field(default=True)
    """Whether to apply test-time augmentation (all 90 degrees rotations and flips)."""

    # Dataloader parameters
    batch_size: int = Field(default=1, ge=1)
    """Batch size for prediction."""

    @field_validator("tile_overlap")
    @classmethod
    def all_elements_non_zero_even(
        cls, tile_overlap: Optional[list[int]]
    ) -> Optional[list[int]]:
        """
        Validate tile overlap.

        Overlaps must be non-zero, positive and even.

        Parameters
        ----------
        tile_overlap : list[int] or None
            Patch size.

        Returns
        -------
        list[int] or None
            Validated tile overlap.

        Raises
        ------
        ValueError
            If the patch size is 0.
        ValueError
            If the patch size is not even.
        """
        if tile_overlap is not None:
            for dim in tile_overlap:
                if dim < 1:
                    raise ValueError(
                        f"Patch size must be non-zero positive (got {dim})."
                    )

                if dim % 2 != 0:
                    raise ValueError(f"Patch size must be even (got {dim}).")

        return tile_overlap

    @field_validator("tile_size")
    @classmethod
    def tile_min_8_power_of_2(
        cls, tile_list: Optional[list[int]]
    ) -> Optional[list[int]]:
        """
        Validate that each entry is greater or equal than 8 and a power of 2.

        Parameters
        ----------
        tile_list : list of int
            Patch size.

        Returns
        -------
        list of int
            Validated patch size.

        Raises
        ------
        ValueError
            If the patch size if smaller than 8.
        ValueError
            If the patch size is not a power of 2.
        """
        patch_size_ge_than_8_power_of_2(tile_list)

        return tile_list

    @field_validator("axes")
    @classmethod
    def axes_valid(cls, axes: str) -> str:
        """
        Validate axes.

        Axes must:
        - be a combination of 'STCZYX'
        - not contain duplicates
        - contain at least 2 contiguous axes: X and Y
        - contain at most 4 axes
        - not contain both S and T axes

        Parameters
        ----------
        axes : str
            Axes to validate.

        Returns
        -------
        str
            Validated axes.

        Raises
        ------
        ValueError
            If axes are not valid.
        """
        # Validate axes
        check_axes_validity(axes)

        return axes

    @model_validator(mode="after")
    def validate_dimensions(self: Self) -> Self:
        """
        Validate 2D/3D dimensions between axes and tile size.

        Returns
        -------
        Self
            Validated prediction model.
        """
        expected_len = 3 if "Z" in self.axes else 2

        if self.tile_size is not None and self.tile_overlap is not None:
            if len(self.tile_size) != expected_len:
                raise ValueError(
                    f"Tile size must have {expected_len} dimensions given axes "
                    f"{self.axes} (got {self.tile_size})."
                )

            if len(self.tile_overlap) != expected_len:
                raise ValueError(
                    f"Tile overlap must have {expected_len} dimensions given axes "
                    f"{self.axes} (got {self.tile_overlap})."
                )

            if any((i >= j) for i, j in zip(self.tile_overlap, self.tile_size)):
                raise ValueError("Tile overlap must be smaller than tile size.")

        return self

    @model_validator(mode="after")
    def std_only_with_mean(self: Self) -> Self:
        """
        Check that mean and std are either both None, or both specified.

        Returns
        -------
        Self
            Validated prediction model.

        Raises
        ------
        ValueError
            If std is not None and mean is None.
        """
        # check that mean and std are either both None, or both specified
        if not self.image_means and not self.image_stds:
            raise ValueError("Mean and std must be specified during inference.")

        if (self.image_means and not self.image_stds) or (
            self.image_stds and not self.image_means
        ):
            raise ValueError(
                "Mean and std must be either both None, or both specified."
            )

        elif (self.image_means is not None and self.image_stds is not None) and (
            len(self.image_means) != len(self.image_stds)
        ):
            raise ValueError(
                "Mean and std must be specified for each " "input channel."
            )

        return self

    def _update(self, **kwargs: Any) -> None:
        """
        Update multiple arguments at once.

        Parameters
        ----------
        **kwargs : Any
            Key-value pairs of arguments to update.
        """
        self.__dict__.update(kwargs)
        self.__class__.model_validate(self.__dict__)

    def set_3D(self, axes: str, tile_size: list[int], tile_overlap: list[int]) -> None:
        """
        Set 3D parameters.

        Parameters
        ----------
        axes : str
            Axes.
        tile_size : list of int
            Tile size.
        tile_overlap : list of int
            Tile overlap.
        """
        self._update(axes=axes, tile_size=tile_size, tile_overlap=tile_overlap)
